#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 17:42:56 2022

@author: ghiggi
"""
import os
import glob 
import yaml

# TODO: to improve a lot !!! 
# - Check metadata !!! 
# - Define rules !!!


def read_yaml(fpath):
    with open(fpath, "r") as f:
        attrs = yaml.safe_load(f)
    return attrs

def identify_missing_metadata(metadata_fpaths, keys):
    if isinstance(keys, str): 
        keys = [keys]
    for key in keys:
        for fpath in metadata_fpaths:
           #  print(fpath)
           metadata = read_yaml(fpath)
           if len(str(metadata.get(key,''))) == 0: # ensure is string to avoid error 
               print(f"Missing {key} at: ",fpath)
    return None 

def identify_missing_coords(metadata_fpaths):
    for fpath in metadata_fpaths:
       metadata = read_yaml(fpath)
       longitude = metadata.get('longitude', -9999)
       latitude = metadata.get('latitude', -9999)
       # Check type validity 
       if isinstance(longitude, str): 
           raise TypeError(f"longitude is not defined as numeric at {fpath}.")
       if isinstance(latitude, str): 
           raise TypeError(f"latitude is not defined as numeric at {fpath}.")
       # Check is not none
       if isinstance(longitude, type(None)) or isinstance(latitude, type(None)): 
           print(f"Unspecified lat lon coordinates at: {fpath}")
           continue
       # Check value validity 
       if  longitude == -9999 or latitude == -9999:
           print(f"Missing lat lon coordinates at: {fpath}")
       elif longitude > 180 or longitude < -180:  
           print("Unvalid longitude at : ",fpath)
       elif latitude > 90 or latitude < -90:
           print("Unvalid latitude at : ",fpath)
       else: 
           pass
    return None 

# # EXAMPLE
# ARCHIVE_DIR = "/ltenas3/0_Data/DISDRODB/Raw" # ARCHIVE_DIR = "/ltenas3/0_Data/DISDRODB/Processed"
# metadata_fpaths = glob.glob(os.path.join(ARCHIVE_DIR, "*/*/metadata/*.yml"))
# identify_missing_coords(metadata_fpaths)

# identify_missing_metadata(metadata_fpaths, keys="station_id") 
# identify_missing_metadata(metadata_fpaths, keys="station_name")
# identify_missing_metadata(metadata_fpaths, keys=["station_id","station_name"])
# identify_missing_metadata(metadata_fpaths, keys="campaign_name")

 #-----------------------------------------------------------------------------.
 # TODO 
 # - station_id 
 # - station_number --> to be replaced by station_id 
 # - station_name --> if missing, use station_id 
 
 # - TODO: define dtype of metadata !!!
 # - raise error if missing station_id
 # - raise error if missing campaign_name 
 #  
 




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 17:07:56 2022

@author: ghiggi
"""
import numpy as np
from disdrodb.standards import (
    get_diameter_bin_center,
    get_diameter_bin_lower,
    get_diameter_bin_upper,
    get_diameter_bin_width,
    get_velocity_bin_center,
    get_velocity_bin_lower,
    get_velocity_bin_upper,
    get_velocity_bin_width,
    
    get_description_dict,
    get_units_dict,
    get_long_name_dict,
    get_data_format_dict,
    get_L0B_encodings_dict,
    
)
#------------------------------------------------------------
# TODO: 
# - check variables in L0A_dtypes match L0B_encodings.yml keys

#------------------------------------------------------------

sensor_name = "OTT_Parsivel"  # LITTLE ISSUE WHEN IT STARTS THE DIAMETER ... 
sensor_name = "OTT_Parsivel2" # LITTLE ISSUE WHEN IT STARTS THE DIAMETER ... 
sensor_name = "Thies_LPM"     # OK 

def check_bin_consistency(sensor_name): 
    
    diameter_bin_lower = get_diameter_bin_lower(sensor_name)
    diameter_bin_upper = get_diameter_bin_upper(sensor_name)
    diameter_bin_center = get_diameter_bin_center(sensor_name)
    diameter_bin_width = get_diameter_bin_width(sensor_name)
    diameter_bin_lower = np.array(diameter_bin_lower)
    diameter_bin_upper = np.array(diameter_bin_upper)
    diameter_bin_center = np.array(diameter_bin_center)
    diameter_bin_width = np.array(diameter_bin_width)
    
    np.testing.assert_allclose(diameter_bin_upper - diameter_bin_lower, diameter_bin_width)
    np.testing.assert_allclose(diameter_bin_lower + diameter_bin_width/2, diameter_bin_center)
    np.testing.assert_allclose(diameter_bin_upper - diameter_bin_width/2, diameter_bin_center)
    
    
    velocity_bin_lower = get_velocity_bin_lower(sensor_name) 
    velocity_bin_upper = get_velocity_bin_upper(sensor_name) 
    velocity_bin_center = get_velocity_bin_center(sensor_name) 
    velocity_bin_width = get_velocity_bin_width(sensor_name) 
             
    velocity_bin_lower = np.array(velocity_bin_lower)
    velocity_bin_upper = np.array(velocity_bin_upper)
    velocity_bin_center = np.array(velocity_bin_center)
    velocity_bin_width = np.array(velocity_bin_width)  
    
    np.testing.assert_allclose(velocity_bin_upper - velocity_bin_lower, velocity_bin_width)
    np.testing.assert_allclose(velocity_bin_lower + velocity_bin_width/2, velocity_bin_center)
    np.testing.assert_allclose(velocity_bin_upper - velocity_bin_width/2, velocity_bin_center)

def check_variable_keys_consistency(sensor_name): 
    description_dict = get_description_dict(sensor_name)
    units_dict = get_units_dict(sensor_name)
    long_name_dict = get_long_name_dict(sensor_name)
    data_format_dict = get_data_format_dict(sensor_name)
    encoding_dict = get_L0B_encodings_dict(sensor_name)
    encoding_vars = set(encoding_dict.keys())
    data_format_vars = set(data_format_dict.keys())
    long_name_vars = set(long_name_dict.keys())
    units_vars = set(units_dict.keys())
    description_vars = set(description_dict.keys())
     
    encoding_vars.difference(data_format_vars)
    encoding_vars.difference(units_vars)
    encoding_vars.difference(description_vars)
    # encoding_vars.difference(long_name_vars) # TODO ADD
    
    data_format_vars.difference(encoding_vars)
    units_vars.difference(encoding_vars)
    description_vars.difference(encoding_vars)
    # long_name_vars.difference(encoding_vars) # TODO ADD


 
# get_available_sensor_name()

# get_diameter_bins_dict(sensor_name)
# get_velocity_bins_dict(sensor_name)

# get_variables_dict(sensor_name)
# get_sensor_variables(sensor_name)

# get_units_dict(sensor_name)
# get_explanations_dict(sensor_name)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 23 14:02:12 2022

@author: ghiggi
"""
import dask
import logging
import numpy as np
import xarray as xr

logger = logging.getLogger(__name__)

def ensure_monotonic_dimension(fpaths, list_ds, dim="time"):
    """Ensure that a list of xr.Dataset has a monotonic increasing dimension."""
    # Get dimension values array (and associated list_ds/xr.Dataset indices)    
    dim_values = np.concatenate([ds[dim].values for ds in list_ds])
    list_index = np.concatenate([np.ones(len(ds[dim]))*i for i, ds in enumerate(list_ds)])
    list_index = list_index.astype(int)
    ds_index = np.concatenate([np.arange(0, len(ds[dim])) for i, ds in enumerate(list_ds)])
       
    # Identify index where start decreasing or duplicated
    diff_dim_values = np.diff(dim_values)
    indices_decreasing = np.where(diff_dim_values.astype(float) <= 0)[0] + 1
  
    if len(indices_decreasing) > 0: 
        idx_start_decreasing = indices_decreasing[0] # or duplicate
        
        # Find last timestep that is lower or equal to that timestep
        idx_restart_increase = np.max(np.where(dim_values <= dim_values[idx_start_decreasing-1])[0])
    
        # Indices to remove 
        idx_to_remove = np.arange(idx_start_decreasing, idx_restart_increase+1)
        list_index_bad = list_index[idx_to_remove]
        ds_index_bad = ds_index[idx_to_remove]
        dim_values_bad = dim_values[idx_to_remove]
        
        # Get dictionary of values/idx which are duplicated (or lead to no monotonic dimension) 
        dict_ds_bad_values = {k: dim_values_bad[np.where(list_index_bad==k)[0]] for k in np.unique(list_index_bad)}
        dict_ds_bad_idx = {k: ds_index_bad[np.where(list_index_bad==k)[0]] for k in np.unique(list_index_bad)}
        
        # Print message  
        for ds_idx_bad, bad_values in dict_ds_bad_values.items(): 
            fpath = fpaths[ds_idx_bad]
            msg = f"In {fpath}, dropping {dim} values {bad_values} to ensure monotonic {dim} dimension."
            logger.debug(msg)
        
        # Remove non-unique and not  da 
        for ds_idx_bad, bad_idx in dict_ds_bad_idx.items():
            ds = list_ds[ds_idx_bad]
            list_ds[ds_idx_bad] = ds.drop_isel({dim: bad_idx})
        
        # Iterative check 
        list_ds = ensure_monotonic_dimension(fpaths, list_ds, dim=dim)
        
    # Return list of xr.Dataset with monotonic dimension 
    return list_ds

# ds_index = [0,1,2,3,0,1,2,3,4]
# list_index = [0,0,0,0,1, 1, 1,1, 1]
# dim_values = [0,1,5,5,5, 5, 6,7,8]
# list_index = np.array(list_index)
# dim_values = np.array(dim_values)
# ds_index = np.array(ds_index)
 
def get_common_coords(list_ds): 
    # Get common data vars 
    coords_ref = set(list_ds[0].coords)
    for ds in list_ds:
        coords_ref = set(ds.coords).intersection(coords_ref)
    return coords_ref
        
def get_common_vars(list_ds):
    # Initialize variables in first file 
    vars_init = set(list_ds[0].data_vars)
    n_vars_init = len(vars_init)
    # Initialize common variable reference
    common_vars_ref = set(list_ds[0].data_vars)
    # Initialize dictionary to collect problems 
    dict_problems = {}
    dict_problems['missing_versus_first'] = {}
    dict_problems['additional_versus_first'] = {}
    dict_problems['evolution'] = {}
    dict_problems['missing_versus_first']['ref'] = vars_init
    dict_problems['additional_versus_first']['ref'] = vars_init
    dict_problems['evolution']['ref'] = vars_init
    # Loop over all datasets 
    for i, ds in enumerate(list_ds):
        # Extract current dataset variable info
        vars_ds = set(ds.data_vars)
        n_vars = len(vars_ds)
        # Check if missing variable compared to first 
        vars_intersection_vs_first = vars_ds.intersection(vars_init)
        # Collect information on missing variables (compared to first)
        if len(vars_intersection_vs_first) != n_vars_init: 
            if len(vars_init.difference(vars_ds)) > 0:
                dict_problems['missing_versus_first'][i] = vars_init.difference(vars_ds)
            if len(vars_ds.difference(vars_init)) > 0:    
                dict_problems['additional_versus_first'][i] = vars_ds.difference(vars_init)        
        # Check if missing an additional variable compared to past common variables  
        common_vars_ref_new = vars_ds.intersection(common_vars_ref)
        # If missing, collect information on additional missing variable
        if len(common_vars_ref_new) != len(common_vars_ref): 
             dict_problems['evolution'][i] = common_vars_ref.difference(common_vars_ref_new)   
        
        # Redefine variables common to all datasets 
        common_vars_ref = common_vars_ref_new

    #---------------------------------------------.
    # Return results 
    return common_vars_ref, dict_problems

def get_list_ds(fpaths): 
    @dask.delayed
    def open_dataset_delayed(fpath):
        ds = xr.open_dataset(fpath, chunks={"time": -1}, engine="netcdf4")
        return ds
    
    list_ds_delayed = []
    for fpath in fpaths:
        list_ds_delayed.append(open_dataset_delayed(fpath))
    list_ds = dask.compute(list_ds_delayed)[0]
    return list_ds

def ensure_constant_coords(list_ds, coords):
    dict_problems = {}
    for coord in coords:
        ref_set = set(list_ds[0][coord].values)
        dict_coord = {}
        dict_coord['ref'] = ref_set
        for i, ds in enumerate(list_ds):
            values = set(ds[coord].values)
            if ref_set != values:
                dict_coord[i] = values
                list_ds[i] = ds.assign_coords({coord: list(ref_set)})
        dict_problems[coord] = dict_coord
    return list_ds, dict_problems
                
                
def xr_concat_datasets(fpaths):
    """Concat xr.Dataset in a robust and parallel way.
    
    1. It checks for time dimension monotonicity
    2. It checks by default for a minimum common set of variables 
    Note: client = Client(processes=True) to execute it fast ! 
    """
    #--------------------------------------.
    # Open xr.Dataset lazily in parallel using dask delayed 
    list_ds = get_list_ds(fpaths)
    
    #--------------------------------------.
    # Ensure time dimension is monotonic increasingly 
    list_ds = ensure_monotonic_dimension(fpaths, list_ds, dim="time")
    
    #--------------------------------------.
    # Ensure coordinate outside time do not vary 
    coords_var = get_common_coords(list_ds)
    coords_without_time = coords_var.difference(set(["time","latitude","longitude","lat","lon"]))
    list_ds, dict_problems = ensure_constant_coords(list_ds, coords = coords_without_time)
    #--------------------------------------.
    # - Log the possible problems
    for coord in coords_without_time:
        dict_problem = dict_problems[coord]
        ref = dict_problem.pop("ref")
        n_problems = len(dict_problem) 
        n_files = len(list_ds)
        if len(dict_problem) != 0: 
            msg = f"Difference found in {n_problems}/{n_files} files."
            logger.debug(msg)
            msg = f"In the first file, the coordinate {coord} has values {ref}."
            logger.debug(msg)
            for i, values in dict_problem.items(): 
                fpath = fpaths[i] 
                msg = f"In {fpath}, the coordinate {coord} has values {values}."
                logger.debug(msg)
                logger.info("Such coordinates have been replaced with the one observed in the first file.")
    
    #--------------------------------------.
    # Get set of common_vars  
    common_vars, dict_problems = get_common_vars(list_ds) 
    
    #--------------------------------------.
    # Log the missing variables compared to first file
    dict_problem = dict_problems["missing_versus_first"]
    n_problems = len(dict_problem) 
    n_files = len(list_ds)
    ref = dict_problem.pop("ref")
    if len(dict_problem) != 0: 
        msg = f"Difference found in {n_problems}/{n_files} files."
        for i, vars in dict_problem.items(): 
            fpath = fpaths[i] 
            msg = f"At file {i}/{n_files} ({fpath}), the variables {vars} are missing (compared to first file)."
            logger.debug(msg)
            
    #--------------------------------------.
    # Log the additional variables compared to first file       
    dict_problem = dict_problems["additional_versus_first"]
    n_problems = len(dict_problem) 
    n_files = len(list_ds)
    ref = dict_problem.pop("ref")
    if len(dict_problem) != 0: 
        msg = f"Difference found in {n_problems}/{n_files} files."
        for i, vars in dict_problem.items(): 
            fpath = fpaths[i] 
            msg = f"At file {i}/{n_files} ({fpath}), there are the variables {vars} which are missing in the first file."
            logger.debug(msg)

    #--------------------------------------.        
    # Log the progressively shrinkage of common variables 
    dict_problem = dict_problems["evolution"]
    n_problems = len(dict_problem) 
    n_files = len(list_ds)
    ref = dict_problem.pop("ref")
    if len(dict_problem) != 0: 
        msg = "Here we report the file which led to shrinkage of the set of common variables."
        logger.debug(msg)
        for i, removed_vars in dict_problem.items(): 
            fpath = fpaths[i] 
            for var in removed_vars: 
                msg = f"At file {i}/{n_files} ({fpath}), the {var} is removed from the common set of variables."
                logger.debug(msg)
                
    #--------------------------------------.                    
    # Concat/merge all netCDFs
    # - If there are common variables, use xr.concat 
    if len(common_vars) > 0:
        # Ensure common set of variables across xr.Datasets 
        for i, ds in enumerate(list_ds):
            list_ds[i] = ds[common_vars]
        # Try concatenating 
        try: 
            logger.info("Start concatenating with xr.concat.")
            ds = xr.concat(list_ds, dim="time", 
                            coords="minimal", 
                            compat="override")
            logger.info("Concatenation with xr.concat has been successful.")
        except:  
            msg = "Concatenation with xr.concat failed."
            logger.error(msg)
            raise ValueError(msg)
    # - Otherwise use xr.merge 
    else: 
        try: 
            logger.info("Start concatenating with xr.merge.")
            ds = xr.merge(list_ds,
                          compat="override",
                          join="outer", 
                          combine_attrs="override")
            logger.info("Concatenation with xr.merge has been successful.")
        except:  
            msg = "Concatenation with xr.merge failed."
            logger.error(msg)
            raise ValueError(msg)
    #--------------------------------------.          
    # Return xr.Dataset
    return ds 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan  1 17:18:04 2022

@author: ghiggi
"""
import os
import subprocess
from disdrodb.utils.parser import get_parser_cmd
from pathlib import Path

# You need to set the disdrodb repo path in your .bashrc
# export PYTHONPATH="${PYTHONPATH}:/home/sguzzo/Projects/disdrodb"
# You need to activate the disdrodb environment: conda activate disdrodb

# -----------------------------------------------------------------------------.
#### Define campaign dictionary
DELFT_dict = {
    "CABAUW": "parser_RASPBERRY.py",
}

#### Define filepaths
home = str(Path.home())
parser_dir = "/home/sguzzo/PycharmProjects/disdrodb/disdrodb/readers/DELFT"
raw_base_dir = "/home/sguzzo/Parsivel/RAW_TELEGRAM"
processed_base_dir = "/tmp/Processed"
# processed_base_dir = "/tmp/Processed/DELFT"

#### Processing settings
l0_processing = True
l1_processing = True
force = True
verbose = True
debugging_mode = False # da cambiare se vuoi processare piu dei primi 5 files
lazy = False
# write_zarr = False
write_netcdf = True

#### Process all campaigns
for campaign_name in DELFT_dict.keys():
    parser_filepath = os.path.join(parser_dir, DELFT_dict[campaign_name])
    cmd = get_parser_cmd(
        parser_filepath=parser_filepath,
        raw_dir=os.path.join(raw_base_dir, campaign_name),
        processed_dir=os.path.join(processed_base_dir, campaign_name),
        l0_processing=l0_processing,
        l1_processing=l1_processing,
        write_netcdf=write_netcdf,
        force=force,
        verbose=verbose,
        debugging_mode=debugging_mode,
        lazy=lazy,
    )

    subprocess.run(cmd, shell=True)
    # os.system(cmd)

# -----------------------------------------------------------------------------.
# TODO:
# --> Useful to test changes to code do not crash other parser
# --> debugging_mode=True to speed up tests ;)

# -----------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import logging
import numpy as np
import pandas as pd
import dask.dataframe as dd
from disdrodb.L0.standards import get_data_format_dict, get_L0A_dtype


logger = logging.getLogger(__name__)


def available_sensor_name():
    from disdrodb.L0.standards import get_available_sensor_name
    sensor_list = get_available_sensor_name()
    raise ValueError("This need to be deprecated in favour of get_available_sensor_name() !") # TODO !!!
    return sensor_list


def check_sensor_name(sensor_name):
    from disdrodb.L0.standards import get_available_sensor_name
    available_sensor_name =  get_available_sensor_name()
    if not isinstance(sensor_name, str):
        logger.exception("'sensor_name' must be a string'")
        raise TypeError("'sensor_name' must be a string'")
    if sensor_name not in available_sensor_name:
        msg = f"Valid sensor_name are {available_sensor_name}"
        logger.exception(msg)
        raise ValueError(msg)
    return

def check_L0A_column_names(df, sensor_name):
    "Checks that the dataframe columns respects DISDRODB standards."
    # Get valid columns 
    dtype_dict = get_L0A_dtype(sensor_name)
    valid_columns = list(dtype_dict)
    valid_columns = valid_columns + ['time']
    valid_columns = set(valid_columns)  
    # Get dataframe column names
    df_columns = list(df.columns)
    df_columns = set(df_columns)
    # --------------------------------------------
    # Check there aren't valid columns 
    unvalid_columns = list(df_columns.difference(valid_columns))
    if len(unvalid_columns) > 0:
        msg = f"The following columns do no met the DISDRODB standards: {unvalid_columns}"
        logger.error(msg) 
        raise ValueError(msg)
    # --------------------------------------------
    # Check time column is present 
    if 'time' not in df_columns:
        msg = "The 'time' column is missing in the dataframe."
        logger.error(msg) 
        raise ValueError(msg)
    # --------------------------------------------
    return None


def check_array_lengths_consistency(df, sensor_name, lazy=True, verbose=False):
    from disdrodb.L0.standards import get_raw_field_nbins

    n_bins_dict = get_raw_field_nbins(sensor_name=sensor_name)
    list_unvalid_row_idx = []
    for key, n_bins in n_bins_dict.items():
        # Check key is available in dataframe
        if key not in df.columns:
            continue
        # Parse the string splitting at ,
        df_series = df[key].astype(str).str.split(",")
        # Check all arrays have same length
        if lazy:
            arr_lengths = df_series.apply(len, meta=(key, "int64"))
            arr_lengths = arr_lengths.compute()
        else:
            arr_lengths = df_series.apply(len)
        idx, count = np.unique(arr_lengths, return_counts=True)
        n_max_vals = idx[np.argmax(count)]
        # Idenfity rows with unexpected array length
        unvalid_row_idx = np.where(arr_lengths != n_max_vals)[0]
        if len(unvalid_row_idx) > 0:
            list_unvalid_row_idx.append(unvalid_row_idx)
    # Drop unvalid rows
    unvalid_row_idx = np.unique(list_unvalid_row_idx)
    if len(unvalid_row_idx) > 0:
        if lazy:
            n_partitions = df.npartitions
            df = df.compute()
            df = df.drop(df.index[unvalid_row_idx])
            df = dd.from_pandas(df, npartitions=n_partitions)
        else:
            df = df.drop(df.index[unvalid_row_idx])
    return df


def check_L0A_standards(fpath, sensor_name, raise_errors=False, verbose=True):
    # Read parquet
    df = pd.read_parquet(fpath)
    # -------------------------------------
    # Check data range
    dict_field_value_range = get_field_value_range_dict(sensor_name)
    list_wrong_columns = []
    for column in df.columns:
        if column in list(dict_field_value_range.keys()):
            if dict_field_value_range[column] is not None: 
                if not df[column].between(*dict_field_value_range[column]).all():
                    list_wrong_columns.append(column)
                    if raise_errors:
                        raise ValueError(f"'column' {column} has values outside the expected data range.")

    if verbose:
        if len(list_wrong_columns) > 0:
            print(" - This columns have values outside the expected data range:", list_wrong_columns)
    # -------------------------------------
    # Check categorical data values
    dict_field_values = get_field_value_options_dict(sensor_name)
    list_wrong_columns = []
    list_msg = []
    for column in df.columns:
        if column in list(dict_field_values.keys()):
            if not df[column].isin(dict_field_values[column]).all():
                list_wrong_columns.append(column)
                if raise_errors:
                    msg = f"'column' {column} has values different from {dict_field_values[column]}"
                    list_msg.append(msg)
                    raise ValueError(msg)
    if verbose:
        if len(list_wrong_columns) > 0:
            print(
                " - The following columns have values outside the expected data range:",
                list_wrong_columns,
            )
            [print(msg) for msg in list_msg]
    # -------------------------------------
    # Check if latitude and longitude are columns of the dataframe
    # - They should be only provided if the instrument is moving !!!!
    if "latitude" in df.columns:
        msg = " - The L0A dataframe has column 'latitude'. "
        "This should be included only if the sensor is moving. "
        "Otherwise, specify the 'latitude' in the metadata !"
        print(msg)
        logger.info(msg)

    if "longitude" in df.columns:
        msg = " - The L0A dataframe has column 'longitude'. "
        "This should be included only if the sensor is moving. "
        "Otherwise, specify the 'longitude' in the metadata !"
        print(msg)
        logger.info(msg)

    # -------------------------------------
    # Check if raw spectrum and 1D derivate exists
    list_sprectrum_vars = ["raw_drop_concentration", "raw_drop_average_velocity", "raw_drop_number"]
    unavailable_vars = np.array(list_sprectrum_vars)[
        np.isin(list_sprectrum_vars, df.columns, invert=True)
    ]
    # Also if Thies_LPM has list_sprectrum_vars?
    if len(unavailable_vars) > 0:
        msg = f" - The variables {unavailable_vars} are not present in the L0 dataframe."
        print(msg)
        logger.info(msg)

    # -------------------------------------
    # Check consistency of array lengths
    # TODO
    # df = check_array_lengths_consistency(df, sensor_name, lazy=True, verbose=verbose)

    # -------------------------------------
    # Add index to dataframe
    # TODO

    # -------------------------------------
    # TODO[GG]:
    # if not respect standards, print errors and remove file (before raising errors)
    # - log, verbose ... L0A conforms to DISDRODB standards ;)

    # -------------------------------------
    return


def check_L0B_standards(x):
    # TODO:
    # - Check for realistic values after having removed the flags !!!!
    pass


####--------------------------------------------------------------------------.
#### Get instrument default string standards
def get_field_ndigits_natural_dict(sensor_name):
    """Get number of digits on th left side of the comma."""
    # (example: 123,45 -> 123)
    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_naturals"] for k, v in data_dict.items()}
    return d


def get_field_ndigits_decimals_dict(sensor_name):
    """Get number of digits on the right side of the comma."""
    # (example: 123,45 -> 45)
    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_decimals"] for k, v in data_dict.items()}
    return d


def get_field_ndigits_dict(sensor_name):
    """Get number of digits

    It excludes the comma but it count the minus sign !!!.
    """
    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_digits"] for k, v in data_dict.items()}
    return d


def get_field_nchar_dict(sensor_name):
    """Get the total number of characters.

    It accounts also for the comma and the minus sign.
    """
    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["n_characters"] for k, v in data_dict.items()}
    return d


def get_field_value_range_dict(sensor_name):
    """Get the variable data range (including nan flags)."""
    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["data_range"] for k, v in data_dict.items()}
    return d


def get_field_flag_dict(sensor_name):
    """Get the variable nan flags."""
    data_dict = get_data_format_dict(sensor_name)
    d = {k: v["nan_flags"] for k, v in data_dict.items()}
    return d


# TODO: get_field_value_realistic_range  # when removing flags

####--------------------------------------------------------------------------.
#### Instrument default numeric standards
# TODO: this is currently used?
# --> YAML file with a list of variables providing error infos?
# --> And provide values that represents errors?


def get_field_value_options_dict(sensor_name):
    if sensor_name == "OTT_Parsivel":
        value_dict = {
            "sensor_status": [0, 1, 2, 3],
            "error_code": [0, 1, 2],
            # TODO: weather codes
            # Faculative/custom fields
            # 'datalogger_temperature': ['NaN']
            "datalogger_voltage": ["OK", "KO"],
            "datalogger_error": [0],
        }
    elif sensor_name == "OTT_Parsivel2":
        value_dict = {
            "sensor_status": [0, 1, 2, 3],
            "error_code": [0, 1, 2],
            # TODO: weather codes
            # Faculative/custom fields
        }
    elif sensor_name == "Thies_LPM":
        value_dict = {
            "laser_status": [0, 1],
            "laser_temperature_analog_status": [0, 1],
            "laser_temperature_digital_status": [0, 1],
            "laser_current_analog_status": [0, 1],
            "laser_current_digital_status": [0, 1],
            "sensor_voltage_supply_status": [0, 1],
            "current_heating_pane_transmitter_head_status": [0, 1],
            "current_heating_pane_receiver_head_status": [0, 1],
            "temperature_sensor_status": [0, 1],
            "current_heating_voltage_supply_status": [0, 1],
            "current_heating_house_status": [0, 1],
            "current_heating_heads_status": [0, 1],
            "current_heating_carriers_status": [0, 1],
            "control_output_laser_power_status": [0, 1],
            "reserve_status": [0, 1]
            # TODO: weather codes
            # Faculative/custom fields
        }
    else:
        raise NotImplementedError

    return value_dict


def get_field_error_dict(device):
    if device == "OTT_Parsivel":
        flag_dict = {
            "sensor_status": [1, 2, 3],
            "datalogger_error": [1],
            "error_code": [1, 2],
        }
    return flag_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

import os
import time
import logging

def create_L0_logger(processed_dir, campaign_name, verbose=False): 
    # Define log name 
    logger_name = "LO_" + "parser_" + campaign_name
    # Create logs directory 
    logs_dir = os.path.join(processed_dir, "logs")
    os.makedirs(logs_dir, exist_ok=True)
    # Create logger 
    _create_logger(logs_dir, logger_name)
    # Retrieve logger
    logger = logging.getLogger(campaign_name)
    # Update logger 
    msg = "### Script started ###"
    if verbose:
        print("\n  " + msg + "\n")
    logger.info(msg)
    # Return logger 
    return logger 
    
def _create_logger(log_dir, logger_name): 
    # Define log file filepath
    logger_fname = f'{time.strftime("%d-%m-%Y_%H-%M-%S")}_{logger_name}.log'
    logger_fpath = os.path.join(log_dir, logger_fname)
    # -------------------------------------------------------------------------.
    # Define logger format
    format_type = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    # Define logger level
    level = logging.DEBUG

    # Define logging
    logging.basicConfig(format=format_type, level=level, filename=logger_fpath)
    return None 

def close_logger(logger):
    handlers = logger.handlers[:]
    for handler in handlers:
        handler.close()
        logger.removeHandler(handler)
    return


#----------------------------
def log_debug(logger, msg, verbose=False):
    logger.debug(msg)
    if verbose:
        print(' - ' + msg)
    
    
def log_info(logger, msg, verbose=False):
    logger.info(msg)
    if verbose:
        print(' - ' + msg)


def log_warning(logger, msg, verbose=False):
    logger.warning(msg)
    if verbose:
        print(' - ' + msg)


#----------------------------
# TO DEPRECATE 

def create_logger(log_dir, logger_name):
    # Define log file filepath
    logger_fname = f'{time.strftime("%d-%m-%Y_%H-%M-%S")}_{logger_name}.log'
    logger_fpath = os.path.join(log_dir, logger_fname)
    # -------------------------------------------------------------------------.
    # Define logger format
    format_type = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    # Define logger level
    level = logging.DEBUG

    # Define logging
    logging.basicConfig(format=format_type, level=level, filename=logger_fpath)

    # -------------------------------------------------------------------------.




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:33:39 2022

@author: ghiggi
"""
import numpy as np 
import pandas as pd 
import xarray as xr 
from xarray.core import dtypes 

def regularize_dataset(ds: xr.Dataset, range_freq: str, 
                       tolerance=None, method=None, fill_value=dtypes.NA):
    """Regularize a dataset across time dimension with uniform resolution."""
    start = ds.time.values[0]
    end = ds.time.values[-1]
    # start_date = pd.to_datetime(start).date() # to start at 00 
    # end_date = pd.to_datetime(end).date() + datetime.timedelta(hours=23, minutes=57, seconds=30)
    new_time_index = pd.date_range(start=pd.to_datetime(start), 
                               end=pd.to_datetime(end) , 
                               freq=range_freq) 
    
    # Regularize dataset and fill with NA values 
    ds_reindexed = ds.reindex({"time": new_time_index}, 
                              method=method, # do not fill gaps
                              tolerance=tolerance, # mismatch in seconds 
                              fill_value=fill_value, 
                              )
    return ds_reindexed


 



 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

# File content
# - Functions to convert L0A Apache parquet files to L0B netCDF files

# -----------------------------------------------------------------------------.
import os
import yaml
import logging
import numpy as np
import pandas as pd
import dask.array as da
import xarray as xr

from disdrodb.L0.check_standards import (
    check_sensor_name,
    check_L0B_standards,
    check_array_lengths_consistency,
)
from disdrodb.L0.standards import (
    get_diameter_bin_center,
    get_diameter_bin_lower,
    get_diameter_bin_upper,
    get_diameter_bin_width,
    get_velocity_bin_center,
    get_velocity_bin_lower,
    get_velocity_bin_upper,
    get_velocity_bin_width,
    get_raw_field_nbins,
    get_L0B_encodings_dict, 
)

logger = logging.getLogger(__name__)


def get_drop_concentration(arr):
    # TODO
    logger.info("Computing raw_drop_concentration from raw spectrum.")
    return arr[:, :, 0]


def get_drop_average_velocity(arr):
    # TODO
    logger.info("Computing raw_drop_average_velocity from raw spectrum.")
    return arr[:, 0, :]


def check_L0_raw_fields_available(df, sensor_name):
    n_bins_dict = get_raw_field_nbins(sensor_name=sensor_name)
    raw_vars = np.array(list(n_bins_dict.keys()))
    missing_vars = raw_vars[np.isin(raw_vars, list(df.columns), invert=True)]
    if len(missing_vars) > 0:
        raise ValueError(f"The following L0 raw fields are missing: {missing_vars}")


def convert_L0_raw_fields_arr_flags(arr, key):
    # TODO: raw_drop_concentration and raw_drop_average_velocity --> -9.999, has floating number
    pass
    return arr


def set_raw_fields_arr_dtype(arr, key):
    if key == "raw_drop_number":
        arr = arr.astype(int)
    else:
        arr = arr.astype(float)
    return arr


def reshape_L0_raw_drop_number_matrix_to_2D(arr, n_bins_dict, n_timesteps):
    try:
        arr = arr.reshape(n_timesteps, n_bins_dict["raw_drop_concentration"], n_bins_dict["raw_drop_average_velocity"])
    except Exception as e:
        msg = f"Impossible to reshape the raw_spectrum matrix. The error is: \n {e}"
        logger.error(msg)
        print(msg)
        raise ValueError(msg)
    return arr


def retrieve_L1_raw_arrays(df, sensor_name, lazy=True, verbose=False):
    # Log
    msg = " - Retrieval of L1 data matrix started."
    if verbose:
        print(msg)
    logger.info(msg)
    # ----------------------------------------------------------.
    # Check L0 raw field availability
    # check_L0_raw_fields_available(df, sensor_name)
    # Retrieve raw fields matrix bins dictionary
    n_bins_dict = get_raw_field_nbins(sensor_name=sensor_name)
    # Retrieve number of timesteps
    if lazy:
        n_timesteps = df.shape[0].compute()
    else:
        n_timesteps = df.shape[0]


    if sensor_name in ['OTT_Parsivel', 'OTT_Parsivel2']:
        split_str = ','
        # Found a campaing (MELBOURNE_2007_OTT) with different different divider, this is a temporary solution
        try:
            head = df.head(1)
            head = head['raw_drop_concentration']
            if head.find(',') == -1:
                    split_str = head[6]
        except KeyError:
            msg = "Something wrong with divider for L0B matrix, default divider is ',', tried to parse with {}".format(split_str)
            if verbose:
                print(msg)
            logger.info(msg)
    if sensor_name in ['Thies_LPM']:
        split_str = ';'
    
    # Retrieve available arrays
    dict_data = {}
    unavailable_keys = []
    for key, n_bins in n_bins_dict.items():
        # Check key is available in dataframe
        if key not in df.columns:
            unavailable_keys.append(key)
            continue
        # Parse the string splitting at ,
        df_series = df[key].astype(str).str.split(split_str)
        # Create array
        if lazy:
            arr = da.stack(df_series, axis=0)
        else:
            arr = np.stack(df_series, axis=0)
        # Remove '' at the last array position
        arr = arr[:, 0 : n_bins_dict[key]]
        # Deal with flag values (-9.9999)
        arr = convert_L0_raw_fields_arr_flags(arr, key=key)
        # Set dtype of the matrix
        arr = set_raw_fields_arr_dtype(arr, key=key)
        # For key='raw_drop_number', reshape to 2D matrix
        if key == "raw_drop_number":
            arr = reshape_L0_raw_drop_number_matrix_to_2D(arr, n_bins_dict, n_timesteps)
        # Add array to dictionary
        dict_data[key] = arr

    # Retrieve unavailable keys from raw spectrum
    if len(unavailable_keys) > 0:
        if "raw_drop_number" not in list(dict_data.keys()):
            raise ValueError(
                "The raw spectrum is required to compute unavaible N_D and N_V."
            )
        if "raw_drop_concentration" in unavailable_keys:
            dict_data["raw_drop_concentration"] = get_drop_concentration(dict_data["raw_drop_number"])
        if "raw_drop_average_velocity" in unavailable_keys:
            dict_data["raw_drop_average_velocity"] = get_drop_average_velocity(dict_data["raw_drop_number"])

    # Log
    msg = " - Retrieval of L1 data matrix finished."
    if verbose:
        print(msg)
    logger.info(msg)
    # Return
    return dict_data


def get_coords(sensor_name):
    check_sensor_name(sensor_name=sensor_name)
    coords = {}
    coords["diameter_bin_center"] = get_diameter_bin_center(sensor_name=sensor_name)
    coords["diameter_bin_lower"] = (
        ["diameter_bin_center"],
        get_diameter_bin_lower(sensor_name=sensor_name),
    )
    coords["diameter_bin_upper"] = (
        ["diameter_bin_center"],
        get_diameter_bin_upper(sensor_name=sensor_name),
    )
    coords["diameter_bin_width"] = (
        ["diameter_bin_center"],
        get_diameter_bin_width(sensor_name=sensor_name),
    )
    coords["velocity_bin_center"] = (
        ["velocity_bin_center"],
        get_velocity_bin_center(sensor_name=sensor_name),
    )
    coords["velocity_bin_lower"] = (
        ["velocity_bin_center"],
        get_velocity_bin_lower(sensor_name=sensor_name),
    )
    coords["velocity_bin_upper"] = (
        ["velocity_bin_center"],
        get_velocity_bin_upper(sensor_name=sensor_name),
    )
    coords["velocity_bin_width"] = (
        ["velocity_bin_center"],
        get_velocity_bin_width(sensor_name=sensor_name),
    )
    return coords


def create_L0B_from_L0A(df, attrs, lazy=True, verbose=False):
    # Retrieve sensor name
    sensor_name = attrs["sensor_name"]
    # -----------------------------------------------------------.
    # Preprocess raw_spectrum, diameter and velocity arrays if available
    if np.any(np.isin(["raw_drop_concentration", "raw_drop_average_velocity", "raw_drop_number"], df.columns)):
        # Check dataframe row consistency
        df = check_array_lengths_consistency(
            df, sensor_name=sensor_name, lazy=lazy, verbose=verbose
        )
        # Retrieve raw data matrices
        dict_data = retrieve_L1_raw_arrays(
            df, sensor_name, lazy=lazy, verbose=verbose
        )
        # Define raw data matrix variables for xarray Dataset
        data_vars = {
            "raw_drop_concentration": (["time", "diameter_bin_center"], dict_data["raw_drop_concentration"]),
            "raw_drop_average_velocity": (["time", "velocity_bin_center"], dict_data["raw_drop_average_velocity"]),
            "raw_drop_number": (
                ["time", "diameter_bin_center", "velocity_bin_center"],
                dict_data["raw_drop_number"],
            ),
        }
    else:
        data_vars = {}
    # -----------------------------------------------------------.
    # Define other disdrometer 'auxiliary' variables varying over time dimension
    aux_columns = df.columns[
        np.isin(df.columns, ["raw_drop_concentration", "raw_drop_average_velocity", "raw_drop_number", "time"], invert=True)
    ]
    if lazy:
        aux_data_vars = {
            column: (["time"], df[column].to_dask_array(lengths=True))
            for column in aux_columns
        }
    else:
        aux_data_vars = {
            column: (["time"], df[column].values) for column in aux_columns
        }
    data_vars.update(aux_data_vars)

    # -----------------------------------------------------------.
    # Drop lat/lon array if present (TODO: in future L0 should not contain it)
    if "latitude" in data_vars:
        _ = data_vars.pop("latitude")
    if "longitude" in data_vars:
        _ = data_vars.pop("longitude")
    if "altitude" in data_vars:
        _ = data_vars.pop("altitude")

    # -----------------------------------------------------------.
    # Define coordinates for xarray Dataset
    coords = get_coords(sensor_name=sensor_name)
    coords["time"] = df["time"].values
    coords["crs"] = attrs["crs"]
    if "latitude" in data_vars:
        coords["latitude"] = data_vars["latitude"]
        _ = data_vars.pop("latitude")
    else:
        coords["latitude"] = attrs["latitude"]
    if "longitude" in data_vars:
        coords["longitude"] = data_vars["longitude"]
        _ = data_vars.pop("longitude")
    else:
        coords["longitude"] = attrs["longitude"]
    if "altitude" in data_vars:
        coords["altitude"] = data_vars["altitude"]
        _ = data_vars.pop("altitude")
    else:
        coords["altitude"] = attrs["altitude"]

    # -----------------------------------------------------------
    # Create xarray Dataset
    try:
        ds = xr.Dataset(
            data_vars=data_vars,
            coords=coords,
            attrs=attrs,
        )
    except Exception as e:
        msg = f"Error in the creation of L1 xarray Dataset. The error is: \n {e}"
        logger.error(msg)
        raise ValueError(msg)

    # -----------------------------------------------------------
    # Check L0B standards
    check_L0B_standards(ds)

    # -----------------------------------------------------------
    # TODO: Replace NA flags

    # -----------------------------------------------------------
    return ds


####--------------------------------------------------------------------------.
#### Writers
def sanitize_encodings_dict(encoding_dict, ds):
    for var in ds.data_vars:
        shape = ds[var].shape
        chunks = encoding_dict[var]["chunksizes"]
        if chunks is not None:
            chunks = [
                shape[i] if chunks[i] > shape[i] else chunks[i]
                for i in range(len(chunks))
            ]
            encoding_dict[var]["chunksizes"] = chunks
    return encoding_dict


def rechunk_dataset(ds, encoding_dict):
    for var in ds.data_vars:
        chunks = encoding_dict[var]["chunksizes"]
        if chunks is not None:
            ds[var] = ds[var].chunk(chunks)
    return ds


def write_L0B(ds, fpath, sensor_name):
    # Ensure directory exist 
    os.makedirs(os.path.dirname(fpath), exist_ok=True)
    
    # Get encoding dictionary
    encoding_dict = get_L0B_encodings_dict(sensor_name)
    encoding_dict = {k: encoding_dict[k] for k in ds.data_vars}

    # Ensure chunksize smaller than the array shape)
    encoding_dict = sanitize_encodings_dict(encoding_dict, ds)

    # Rechunk variables for fast writing !
    ds = rechunk_dataset(ds, encoding_dict)

    # Write netcdf
    ds.to_netcdf(fpath, engine="netcdf4", encoding=encoding_dict)


####--------------------------------------------------------------------------.
#### Summary statistics
def create_summary_statistics(ds, processed_dir, station_id, sensor_name):
    """Create L0 summary statistics and save it into the station info YAML file."""
    ###-----------------------------------------------------------------------.
    # Initialize dictionary
    stats_dict = {}
    
    # Infer the sampling interval looking at the difference between timesteps
    dt, counts = np.unique(np.diff(ds.time.values), return_counts=True)
    dt_most_frequent = dt[np.argmax(counts)]
    dt_most_frequent = dt_most_frequent.astype('m8[s]')
    inferred_sampling_interval = dt_most_frequent.astype(int)
    stats_dict['inferred_sampling_interval'] = inferred_sampling_interval

    # Number of years, months, days, minutes 
    time = ds.time.values
    n_timesteps = len(time)
    n_minutes = inferred_sampling_interval/60*n_timesteps
    n_hours = n_minutes/60 
    n_days = n_hours/24

    stats_dict['n_timesteps'] = n_timesteps
    stats_dict['n_minutes'] = n_minutes
    stats_dict['n_hours'] = n_hours
    stats_dict['n_days'] = n_days

    # Add start_time and end_time
    start_time = pd.DatetimeIndex(time[[0]])
    end_time = pd.DatetimeIndex(time[[-1]])
    years = np.unique([start_time.year, end_time.year])
    if len(years) == 1:
        years_coverage = str(years[0])
    else: 
        years_coverage = str(years[0]) + "-" + str(years[-1])
        
    stats_dict['years_coverage'] = years_coverage
    stats_dict['start_time'] = start_time[0].isoformat()
    stats_dict['end_time'] = end_time[0].isoformat()

    ###-----------------------------------------------------------------------.
    # TODO: Create and save image with temporal coverage    
    # --> Colored using quality flag from sensor_status if available ? 
    
    ###-----------------------------------------------------------------------.
    # TODO STATISTICS 
    # --> Requiring deriving stats from raw spectrum

    # diameter_min, diameter_max, diameter_sum
    
    # Total rain events 

    # Total rainy minutes 

    # Total dry minutes 

    # Number of dry/rainy minutes 

    ###-----------------------------------------------------------------------.
    # Save to info.yaml 
    info_path = os.path.join(processed_dir, "info", station_id + ".yml")
    with open(info_path, "w") as f:
         yaml.dump(stats_dict, f, sort_keys=False)
         
    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.
import logging
import os
import shutil
import glob
import numpy as np
import pandas as pd
import dask.dataframe as dd
from disdrodb.utils.logger import log_info, log_warning
logger = logging.getLogger(__name__)

####---------------------------------------------------------------------------.
#### Directory/Filepaths Defaults
def infer_institute_from_fpath(fpath):
    idx_start = fpath.rfind("DISDRODB")
    disdrodb_fpath = fpath[idx_start:]
    institute = disdrodb_fpath.split("/")[2]   
    return institute 

def infer_campaign_from_fpath(fpath):
    idx_start = fpath.rfind("DISDRODB")
    disdrodb_fpath = fpath[idx_start:]
    campaign = disdrodb_fpath.split("/")[3] 
    return campaign 

def infer_station_id_from_fpath(fpath):
    idx_start = fpath.rfind("DISDRODB")
    disdrodb_fpath = fpath[idx_start:]
    station_id = disdrodb_fpath.split("/")[5]
    # Optional strip .yml if fpath point to YAML file 
    station_id.strip(".yml")  
    return station_id 


def get_campaign_name(base_dir):
    """Return the campaign name from 'raw_dir' or processed_dir' paths."""
    base_dir = parse_fpath(base_dir)
    campaign_name = os.path.basename(base_dir).upper()
    return campaign_name

def get_L0A_dir(processed_dir, station_id): 
    dir_path = os.path.join(processed_dir, "L0A", station_id)
    return dir_path

def get_L0A_fname(campaign_name, station_id, suffix=""):
    if suffix != "":
        suffix = "_" + suffix
    fname = campaign_name + "_s" + station_id + suffix + ".parquet"
    return fname


def get_L0A_fpath(processed_dir, station_id, suffix=""):
    campaign_name = get_campaign_name(processed_dir)
    fname = get_L0A_fname(campaign_name, station_id, suffix=suffix)
    dir_path = get_L0A_dir(processed_dir, station_id)
    fpath = os.path.join(dir_path, fname)
    return fpath

def get_L0B_dir(processed_dir, station_id): 
    dir_path = os.path.join(processed_dir, "L0B", station_id)
    return dir_path

def get_L0B_fname(campaign_name, station_id, suffix=""):
    if suffix != "":
        suffix = "_" + suffix
    # TODO: _s make sense with station_id... but if station_name a bit orrible
    fname = campaign_name + "_s" + station_id + suffix + ".nc"
    return fname


def get_L0B_fpath(processed_dir, station_id, suffix=""):
    campaign_name = get_campaign_name(processed_dir)
    fname = get_L0B_fname(campaign_name, station_id, suffix=suffix)
    dir_path = get_L0B_dir(processed_dir, station_id)
    fpath = os.path.join(dir_path, fname)
    return fpath

####--------------------------------------------------------------------------.
#### Directory/File Creation/Deletion


def _create_directory(path):
    if not isinstance(path, str):
        raise TypeError("'path' must be a strig.")
    try:
        os.makedirs(path)
        logger.debug(f"Created directory {path}.")
    except FileExistsError:
        logger.debug(f"Directory {path} already exists.")
        pass
    except (Exception) as e:
        dir_name = os.path.basename(path)
        msg = f"Can not create folder {dir_name} inside <path>. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)


def _remove_if_exists(fpath, force=False):
    if os.path.exists(fpath):
        if not force:
            msg = f"--force is False and a file already exists at:{fpath}"
            logger.error(msg)
            raise ValueError(msg)
        try:
            os.remove(fpath)
        except IsADirectoryError:
            try:
                os.rmdir(fpath)
            except OSError:
                try:
                    # shutil.rmtree(fpath.rpartition('.')[0])
                    for f in glob.glob(fpath + "/*"):
                        try:
                            os.remove(f)
                        except OSError as e:
                            msg = f"Can not delete file {f}, error: {e.strerror}"
                            logger.exception(msg)
                    os.rmdir(fpath)
                except:
                    msg = f"Something wrong with: {fpath}"
                    logger.error(msg)
                    raise ValueError(msg)
        logger.info(f"Deleted folder {fpath}")


####--------------------------------------------------------------------------.
#### Directory checks
def parse_fpath(fpath):
    """Ensure fpath does not end with /."""
    if not isinstance(fpath, str):
        raise TypeError("'parse_fpath' expects a directory/filepath string.")
    if fpath[-1] == "/":
        print("{} should not end with /.".format(fpath))
        fpath = fpath[:-1]
    return fpath

####--------------------------------------------------------------------------.
#### L0 processing directory checks
def check_raw_dir(raw_dir, verbose=False):
    """Check validity of raw_dir.

    Steps:
    1. Check that 'raw_dir' is a valid directory path
    2. Check that 'raw_dir' follows the expect directory structure
    3. Check that each station_id directory contains data
    4. Check that for each station_id the mandatory metadata are specified.

    """
    from disdrodb.L0.metadata import create_metadata
    from disdrodb.L0.metadata import check_metadata_compliance
    from disdrodb.L0.issue import create_issue_yml
    from disdrodb.L0.issue import check_issue_compliance
    # -------------------------------------------------------------------------.
    # Check input argument
    if not isinstance(raw_dir, str):
        raise TypeError("Provide 'raw_dir' as a string'.")
    if not os.path.exists(raw_dir):
        raise ValueError("'raw_dir' {} directory does not exist.".format(raw_dir))
    if not os.path.isdir(raw_dir):
        raise ValueError("'raw_dir' {} is not a directory.".format(raw_dir))
    # -------------------------------------------------------------------------.
    #### Check there is /data subfolders
    list_subfolders = os.listdir(raw_dir)
    if len(list_subfolders) == 0:
        raise ValueError("There are not subfolders in {}".format(raw_dir))
    if "data" not in list_subfolders:
        raise ValueError(
            "'raw_dir' {} should have the /data subfolder.".format(raw_dir)
        )
    
    # -------------------------------------------------------------------------.
    #### Check there are subfolders corresponding to station to process
    raw_data_dir= os.path.join(raw_dir, "data")
    list_data_station_id = os.listdir(raw_data_dir)
    if len(list_data_station_id) == 0:
        raise ValueError("No station directories within {}".format(raw_data_dir))
    
    # -------------------------------------------------------------------------.
    #### Check there are data files in each list_data_station_id
    list_raw_data_station_dir= [
        os.path.join(raw_data_dir, station_id) for station_id in list_data_station_id
    ]
    list_nfiles_per_station = [
        len(glob.glob(os.path.join(path, "*"))) for path in list_raw_data_station_dir
    ]
    idx_0_files = np.where(np.array(list_nfiles_per_station) == 0)[0]
    if len(idx_0_files) > 0:
        empty_station_dir = [list_raw_data_station_dir[idx] for idx in idx_0_files]
        raise ValueError(
            "The following data directories are empty: {}".format(empty_station_dir)
        )
        
    # -------------------------------------------------------------------------.
    #### Check there is /metadata subfolders
    metadata_dir = os.path.join(raw_dir, "metadata")
    if "metadata" not in list_subfolders:
        # - Create metadata directory 
        _create_directory(metadata_dir)
        # - Create default metadata yml file for each station (since the folder didn't existed)
        list_metadata_fpath = [
            os.path.join(metadata_dir, station_id + ".yml")
            for station_id in list_data_station_id
        ]
        _ = [create_metadata(fpath) for fpath in list_metadata_fpath]
        msg = "'raw_dir' {} should have the /metadata subfolder. ".format(raw_dir)
        msg1 = "It has been now created with also empty metadata files to be filled for each station."
        raise ValueError(msg + msg1)

    # -------------------------------------------------------------------------.
    #### Check there are metadata file for each station_id in /metadata
    list_metadata_fpath = glob.glob(os.path.join(metadata_dir, "*.yml"))
    list_metadata_fname = [os.path.basename(fpath) for fpath in list_metadata_fpath]
    list_metadata_station_id = [fname[:-4] for fname in list_metadata_fname]
    
    # - Check there is metadata for each station
    missing_data_station_idx = np.where(
        np.isin(list_data_station_id, list_metadata_station_id, invert=True)
    )[0]
    # - If missing, create the defaults files and raise an error
    if len(missing_data_station_idx) > 0:
        list_missing_station_id = [
            list_data_station_id[idx] for idx in missing_data_station_idx
        ]
        list_missing_metadata_fpath = [
            os.path.join(metadata_dir, station_id + ".yml")
            for station_id in list_missing_station_id
        ]
        _ = [create_metadata(fpath) for fpath in list_missing_metadata_fpath]
        msg = "The metadata files for the following station_id were missing: {}".format(
            list_missing_station_id
        )
        raise ValueError(msg + " Now have been created to be filled.")
        
    # - Check not excess metadata compared to present stations
    excess_metadata_station_idx = np.where(np.isin(list_metadata_station_id, list_data_station_id, invert=True))[0]
    if len(excess_metadata_station_idx) > 0:
        list_excess_station_id = [
            list_metadata_station_id[idx] for idx in excess_metadata_station_idx
        ]
        print(
            "There are the following metadata files without corresponding data: {}".format(
                list_excess_station_id
            )
        )
        
    # -------------------------------------------------------------------------.
    #### Check metadata compliance
    _ = [check_metadata_compliance(fpath) for fpath in list_metadata_fpath]
    # TODO: MISSING IMPLEMENTATION OF check_metadata_compliance
    # -------------------------------------------------------------------------.
    #### Check there is /issue subfolder
    issue_dir = os.path.join(raw_dir, "issue")
    if "issue" not in list_subfolders:
        # - Create issue directory 
        _create_directory(issue_dir)
        # - Create issue yml file for each station (since the folder didn't existed)
        list_issue_fpath = [os.path.join(issue_dir, station_id + ".yml")
            for station_id in list_data_station_id
        ]
        _ = [create_issue_yml(fpath) for fpath in list_issue_fpath]
        msg = "The /issue subfolder has been now created to document and then remove timesteps with problematic data."
        logger.info(msg)
    # -------------------------------------------------------------------------.
    #### Check there are issue file for each station_id in /issue 
    list_issue_fpath = glob.glob(os.path.join(issue_dir, "*.yml"))
    list_issue_fname = [os.path.basename(fpath) for fpath in list_issue_fpath]
    list_issue_station_id = [fname[:-4] for fname in list_issue_fname]
    
    # - Check there is issue for each station
    missing_data_station_idx = np.where(
        np.isin(list_data_station_id, list_issue_station_id, invert=True)
    )[0]
    # - If missing, create the defaults files and raise an error
    if len(missing_data_station_idx) > 0:
        list_missing_station_id = [
            list_data_station_id[idx] for idx in missing_data_station_idx
        ]
        list_missing_issue_fpath = [
            os.path.join(issue_dir, station_id + ".yml")
            for station_id in list_missing_station_id
        ]
        _ = [create_issue_yml(fpath) for fpath in list_missing_issue_fpath]
        msg = "The issue files for the following station_id were missing: {}".format(
            list_missing_station_id
        )
        
    # - Check not excess issue compared to present stations
    excess_issue_station_idx = np.where(np.isin(list_issue_station_id, list_data_station_id, invert=True))[0]
    if len(excess_issue_station_idx) > 0:
        list_excess_station_id = [list_issue_station_id[idx] for idx in excess_issue_station_idx]
        msg =  f"There are the following issue files without corresponding data: {list_excess_station_id}"
        log_warning(logger, msg, verbose)

    # -------------------------------------------------------------------------.
    #### Check issue compliance
    _ = [check_issue_compliance(fpath) for fpath in list_issue_fpath]
    # TODO: MISSING IMPLEMENTATION OF check_issue_compliance
    # -------------------------------------------------------------------------.
    return None
    
def check_processed_dir(processed_dir, force=False):
    """Check that 'processed_dir' is a valid directory path."""
    if not isinstance(processed_dir, str):
        raise TypeError("Provide 'processed_dir' as a string'.")
    #------------------------------    
    # Check processed_dir has "DISDRODB/Processed" to avoid deleting precious stuffs 
    if processed_dir.find("DISDRODB/Processed") == -1: 
        msg = "Expecting 'processed_dir' to contain the pattern */DISDRODB/Processed/*."
        logger.error(msg)
        raise ValueError(msg)
        
    # Check processed_dir does not end with "DISDRODB/Processed" 
    # - It must contain also the <campaign_name> directory  
    if processed_dir.endswith("Processed") or processed_dir.endswith("Processed/"):
        msg = "Expecting 'processed_dir' to contain the pattern */DISDRODB/Processed/<campaign_name>."
        logger.error(msg)
        raise ValueError(msg)

    #------------------------------
    # If forcing overwriting 
    if force: 
        # Check processed_dir is a directory before removing the content 
        if os.path.exists(processed_dir):
            if not os.path.isdir(processed_dir):
                msg = f"'processed_dir' {processed_dir} already exist but is not a directory."
                logger.error(msg)
                raise ValueError(msg) 
            # Remove content of existing processed_dir
            # TODO: maybe remove also the campaign name directory, not only the content of it
            shutil.rmtree(processed_dir)  # TODO: !! TOO DANGEROUS ??? !!!

    #------------------------------
    # If avoiding overwriting 
    if not force:
        if os.path.exists(processed_dir):
            msg = f"'processed_dir' {processed_dir} already exists and force=False."
            logger.error(msg)
            raise ValueError(msg)
            
    #------------------------------        
    # Recreate processed_dir
    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)
    else:
        msg = "Please report the BUG. This should not happen."  
        logger.error(msg)
        raise ValueError(msg)
           

def check_campaign_name(raw_dir, processed_dir):
    """Check that 'raw_dir' and 'processed_dir' have same campaign_name."""
    upper_campaign_name = os.path.basename(raw_dir).upper()
    raw_campaign_name = os.path.basename(raw_dir)
    processed_campaign_name = os.path.basename(processed_dir)
    if raw_campaign_name != processed_campaign_name:
        msg = f"'raw_dir' and 'processed_dir' must ends with same <campaign_name> {upper_campaign_name}"
        logger.error(msg)
        raise ValueError(msg)
            
    if raw_campaign_name != upper_campaign_name:
        msg = f"'raw_dir' and 'processed_dir' must ends with UPPERCASE <campaign_name> {upper_campaign_name}"
        logger.error(msg)
        raise ValueError(msg)

    return upper_campaign_name


def check_directories(raw_dir, processed_dir, force=False):
    """Check that the specified directories respect the standards."""
    raw_dir = parse_fpath(raw_dir)
    processed_dir = parse_fpath(processed_dir)
    check_raw_dir(raw_dir)
    check_processed_dir(processed_dir, force=force)
    check_campaign_name(raw_dir, processed_dir)
    return raw_dir, processed_dir

def check_metadata_dir(processed_path):
    # Create metadata folder
    try:
        metadata_folder_path = os.path.join(processed_path, "metadata")
        os.makedirs(metadata_folder_path)
        logger.debug(f"Created {metadata_folder_path}.")
    except FileExistsError:
        logger.debug(f"{metadata_folder_path} already existed.")
        pass
    except (Exception) as e:
        msg = f"Folder metadata can not be created in {metadata_folder_path}>. \n The error is: {e}."
        logger.exception(msg)
        raise FileNotFoundError(msg)

def copy_metadata_from_raw_dir(raw_dir, processed_dir):
    '''Copy yaml files in raw_dir/metadata into processed_dir/metadata'''
    # Get src and dst metadata directory 
    raw_metadata_dir = os.path.join(raw_dir, "metadata")
    processed_metadata_dir = os.path.join(processed_dir, "metadata")
    # Retrieve metadata fpaths in raw directory
    raw_metadata_fpaths = glob.glob(os.path.join(raw_metadata_dir, "*.yml"))
    # Copy all metadata yml files into the "processed" folder
    for raw_metadata_fpath in raw_metadata_fpaths:
        # Check if is a files
        if os.path.isfile(raw_metadata_fpath):
            metadata_fname = os.path.basename(raw_metadata_fpath)
            processed_metadata_fpath = os.path.join(processed_metadata_dir, metadata_fname)
            try:
                # Copy every file
                shutil.copy(raw_metadata_fpath, processed_metadata_fpath)
                msg = f'{metadata_fname} copied into {processed_metadata_dir}.' 
                logger.info(msg)
            except (Exception) as e:
                msg = f'Something went wrong when copying {metadata_fname} into {processed_metadata_dir}.\n The error is: {e}.'
                logger.error(msg)
        else:
            msg = f'Cannot copy {metadata_fname} into {processed_metadata_dir}.'
            logger.error(msg)
            raise ValueError(msg)
    metadata_fnames = [os.path.basename(metadata_fpath) for metadata_fpath in raw_metadata_fpaths]
    msg = f'The metadata of stations ({metadata_fnames}) have been copied into {processed_metadata_dir}.' 
    logger.info(msg)
    
####--------------------------------------------------------------------------.
#### L0 processing directory structure


def create_directory_structure(raw_dir, processed_dir):
    """Create directory structure for L0A and L0B processing."""
    #-----------------------------------------------------.
    #### Create metadata folder inside processed_dir
    check_metadata_dir(processed_dir)
    copy_metadata_from_raw_dir(raw_dir, processed_dir)
    
    #-----------------------------------------------------.
    #### Create info folder inside processed_dir
    try:
        info_folder_path = os.path.join(processed_dir, "info")
        os.makedirs(info_folder_path)
        logger.debug(f"Created {info_folder_path}")
    except FileExistsError:
        logger.debug(f"Found {info_folder_path}")
        pass
    except (Exception) as e:
        msg = f"Can not create folder metadata inside <info_folder_path>. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)
        
    #-----------------------------------------------------.
    #### Create L0A folder inside processed_dir
    try:
        L0A_folder_path = os.path.join(processed_dir, "L0A")
        os.makedirs(L0A_folder_path)
        logger.debug(f"Created {L0A_folder_path}")
    except FileExistsError:
        logger.debug(f"Found {L0A_folder_path}")
        pass
    except (Exception) as e:
        msg = f"Can not create folder L0A inside {processed_dir}. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)
        
    #-----------------------------------------------------.
    #### Create L0B folder inside processed_dir
    try:
        L0B_folder_path = os.path.join(processed_dir, "L0B")
        os.makedirs(L0B_folder_path)
        logger.debug(f"Created {L0B_folder_path}")
    except FileExistsError:
        logger.debug(f"Found {L0B_folder_path}")
        pass
    except (Exception) as e:
        msg = f"Can not create folder L0B inside {processed_dir}. Error: {e}"
        logger.exception(msg)
        raise FileNotFoundError(msg)
    return


####--------------------------------------------------------------------------.
#### DISDRODB L0A Readers
def _read_L0A(fpath, lazy=True, verbose=False): 
    # Log
    msg = f" - Reading L0 Apache Parquet file at {fpath} started."
    log_info(logger, msg, verbose)
    # Read
    if lazy:
        df = dd.read_parquet(fpath)
    else:
        df = pd.read_parquet(fpath)
    # Log
    msg = f" - Reading L0 Apache Parquet file at {fpath} ended."
    log_info(logger, msg, verbose)
    return df 

def read_L0A_dataframe(fpaths, lazy=True, verbose=False, debugging_mode=False):
    """
    Read DISDRODB L0A Apache Parquet file(s). 

    Parameters
    ----------
    fpaths : str or list
        Either a list or a single filepath .
    verbose : bool
        Whether to print detailed processing information into terminal. 
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        If fpaths is a list, it reads only the first 3 files 
        The default is False.
    lazy : bool
        Whether to read the dataframe lazily with dask. 
        If lazy=True, it returns a dask.dataframe.
        If lazy=False, it returns a pandas.DataFrame  
        The default is True.
    """
    from disdrodb.L0.L0A_processing import concatenate_dataframe
    # ----------------------------------------
    # Check fpaths validity
    if not isinstance(fpaths, (list,str)): 
        raise TypeError("Expecting fpaths to be a string or a list of strings.")
    # TODO:
    # - CHECK ENDS WITH .parquet  
    # ----------------------------------------
    # If list of length 1, convert to string
    if isinstance(fpaths, list) and len(fpaths) == 1:
        fpaths = fpaths[0]
    # ---------------------------------------------------
    # If more than 1 fpath, read and concantenate first
    if isinstance(fpaths, list): 
        if debugging_mode: 
            fpaths = fpaths[0:3] # select first 3 fpaths
        list_df = [_read_L0A(fpath, lazy=lazy, verbose=verbose) for fpath in fpaths]
        df = concatenate_dataframe(list_df, verbose=verbose, lazy=lazy)
    # Else read the single file 
    else: 
        df = _read_L0A(fpaths, lazy=lazy, verbose=verbose)
    # ---------------------------------------------------
    # Return dataframe 
    return df 

####---------------------------------------------------------------------------.
#### TO BE DEPRECATED 
#### Back-compatibility stuffs 
# TODO: DEPRECATE 
def get_L0_fname(campaign_name, station_id, suffix=""):
    return get_L0A_fname(campaign_name, station_id, suffix)


def get_L0_fpath(processed_dir, station_id, suffix=""):
   return get_L0A_fpath(processed_dir, station_id, suffix)


def get_L1_netcdf_fname(campaign_name, station_id, suffix=""):
    return get_L0B_fname(campaign_name, station_id, suffix)


def get_L1_netcdf_fpath(processed_dir, station_id, suffix=""):
    return get_L0B_fpath(processed_dir, station_id, suffix)


def check_L0_is_available(processed_dir, station_id, suffix=""):
    fpath = get_L0A_fpath(processed_dir, station_id, suffix=suffix)
    if not os.path.exists(fpath):
        msg = f"Need to run L0 processing first. The Apache Parquet file {fpath} is not available."
        logger.exception(msg)
        raise ValueError(msg)
    # Log
    msg = f"Found parquet file: {fpath}"
    logger.info(msg)


def read_L0_data(processed_dir, station_id, suffix="", 
                 lazy=True, verbose=False, debugging_mode=False):
    """Read L0 Apache Parquet into dataframe.
    
    If debugging_mode = True, return just a subset of total rows.
    """
    # Check L0 is available
    check_L0_is_available(processed_dir, station_id, suffix=suffix)
    # Define fpath
    fpath = get_L0A_fpath(processed_dir, station_id, suffix=suffix)
    # Read L0A Apache Parquet file 
    df = _read_L0A(fpath, lazy=lazy, verbose=verbose)
    # Subset dataframe if debugging_mode = True
    if debugging_mode:
        if not lazy:
            df = df.iloc[0:100, :]
        else:
            NotImplementedError
    return df


####--------------------------------------------------------------------------.




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

# Retrieve sensor standards and configs

# -----------------------------------------------------------------------------.
import os
import yaml
import logging
import datetime 
    
logger = logging.getLogger(__name__)

PRODUCT_VERSION = "V0"
SOFTWARE_VERSION = "V0"

# TODO: 
# - get_L0_dtype,
# - get_L1_netcdfencoding_dict --> get_encodings(sensor_name)
# - _get_encodings_key
# - get_encodings_dtype 
# - get_encodings_chunk 
# - get_encodings_ ....

def read_config_yml(sensor_name, filename):
    """Read a config yaml file and return the dictionary."""
    # Get config path
    config_sensor_dir_path = get_configs_dir(sensor_name)
    fpath = os.path.join(config_sensor_dir_path, filename)
    # Check yaml file exists
    if not os.path.exists(fpath):
        msg = f"{filename} not available in {config_sensor_dir_path}"
        logger.exception(msg)
        raise ValueError(msg)
    # Open dictionary
    with open(fpath, "r") as f:
        d = yaml.safe_load(f)
    return d


def get_configs_dir(sensor_name):
    """Retrieve configs directory."""
    dir_path = os.path.dirname(__file__)
    config_dir_path = os.path.join(dir_path, "configs")
    config_sensor_dir_path = os.path.join(config_dir_path, sensor_name)
    if not os.path.exists(config_sensor_dir_path):
        list_sensors = sorted(os.listdir(config_dir_path))
        print(f"Available sensor_name are {list_sensors}")
        raise ValueError(
            f"The config directory {config_sensor_dir_path} does not exist."
        )
    return config_sensor_dir_path


def get_available_sensor_name():
    """Get available sensor_name."""
    dir_path = os.path.dirname(__file__)
    config_dir_path = os.path.join(dir_path, "configs")
    # TODO: here add checks that contains all required yaml file
    return sorted(os.listdir(config_dir_path))


def get_variables_dict(sensor_name):
    """Get a dictionary containing the variable name of the sensor field numbers."""
    return read_config_yml(sensor_name=sensor_name, filename="variables.yml")


def get_sensor_variables(sensor_name):
    """Get sensor variable names list."""
    return list(get_variables_dict(sensor_name).values())


def get_data_format_dict(sensor_name):
    """Get a dictionary containing the data format of each sensor variable."""
    return read_config_yml(sensor_name=sensor_name, filename="L0_data_format.yml")

def get_long_name_dict(sensor_name): 
    """Get a dictionary containing the long name of each sensor variable."""
    return read_config_yml(sensor_name=sensor_name, filename="variable_longname.yml")
  
def get_units_dict(sensor_name):
    """Get a dictionary containing the unit of each sensor variable."""
    return read_config_yml(sensor_name=sensor_name, filename="variable_units.yml")


def get_description_dict(sensor_name):
    """Get a dictionary containing the description of each sensor variable."""
    d = read_config_yml(sensor_name=sensor_name, filename="variable_description.yml")
    return d


def get_diameter_bins_dict(sensor_name):
    """Get dictionary with sensor_name diameter bins information."""
    d = read_config_yml(sensor_name=sensor_name, filename="diameter_bins.yml")
    # TODO: Check dict contains center, bounds and width keys
    return d


def get_velocity_bins_dict(sensor_name):
    """Get velocity with sensor_name diameter bins information."""
    d = read_config_yml(sensor_name=sensor_name, filename="velocity_bins.yml")
    return d


def get_L0A_dtype(sensor_name):
    """Get a dictionary containing the L0 dtype."""
    # Note: This function could extract the info from get_L0A_encodings_dict in future.
    d = read_config_yml(sensor_name=sensor_name, filename="L0A_encodings.yml")
    return d

def get_L0A_encodings_dict(sensor_name):
    """Get a dictionary containing the L0 dtype."""
    # L0A_encodings currently specify only the dtype ... could be expanded in future.
    d = read_config_yml(sensor_name=sensor_name, filename="L0A_encodings.yml")
    return d

def get_L0B_encodings_dict(sensor_name):
    """Get a dictionary containing the encoding to write L0B netCDFs."""
    d = read_config_yml(sensor_name=sensor_name, filename="L0B_encodings.yml")

    # Ensure chunksize is a list
    for var in d.keys():
        if not isinstance(d[var]["chunksizes"], (list, type(None))):
            d[var]["chunksizes"] = [d[var]["chunksizes"]]

    # Sanitize encodings
    for var in d.keys():
        # Ensure contiguous=True if chunksizes is None
        if isinstance(d[var]["chunksizes"], type(None)) and not d[var]["contiguous"]:
            # These changes are required to enable netCDF writing
            d[var]["contiguous"] = True
            d[var]["fletcher32"] = False
            d[var]["zlib"] = False
            print(f"Set contiguous=True for variable {var} because chunksizes=None")
            print(f"Set fletcher32=False for variable {var} because contiguous=True")
            print(f"Set zlib=False for variable {var} because contiguous=True")
        # Ensure contiguous=False if chunksizes is not None
        if d[var]["contiguous"] and not isinstance(d[var]["chunksizes"], type(None)):
            d[var]["contiguous"] = False
            print(
                f"Set contiguous=False for variable {var} because chunksizes is defined!"
            )

    return d


def set_DISDRODB_L0_attrs(ds, attrs):
    sensor_name = attrs['sensor_name']
    #----------------------------------
    # Set global attributes (from metadata)
    ds.attrs = attrs
    
    #----------------------------------
    # Set variable attributes (from config standards)
    description_dict = get_description_dict(sensor_name)
    units_dict = get_units_dict(sensor_name)
    long_name_dict = get_long_name_dict(sensor_name)
    data_format_dict = get_data_format_dict(sensor_name)
    for var in list(ds.data_vars):
        attrs_var = {}
        attrs_var['long_name'] = long_name_dict[var] 
        attrs_var['description'] = description_dict[var]
        attrs_var['unit'] = units_dict[var]
        # TODO: 
        # attrs['valid_min'] =  data_format_dict
        # attrs['valid_max'] =  data_format_dict
        ds[var].attrs = attrs_var
        
    #----------------------------------  
    # Set coordinate attributes 
    # TODO 
    # lat/lon - degrees N/E... 
    # altitude - meters above sea level
    
    #----------------------------------
    # Add DISDRODB processing info 
    now = datetime.datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    ds.attrs['disdrodb_processing_date'] = current_time
    ds.attrs['disdrodb_product_version'] = PRODUCT_VERSION 
    ds.attrs['disdrodb_software_version'] = SOFTWARE_VERSION 
    ds.attrs['disdrodb_product_level'] = "L0" 
    # ds.attrs['disdrodb_id'] = # TODO XXXX 
    
    return ds
   
####-------------------------------------------------------------------------.
#############################################
#### Get diameter and velocity bins info ####
#############################################

def get_diameter_bin_center(sensor_name):
    """Get diameter bin center."""
    diameter_dict = get_diameter_bins_dict(sensor_name)
    diameter_bin_center = list(diameter_dict["center"].values())
    return diameter_bin_center


def get_diameter_bin_lower(sensor_name):
    """Get diameter bin lower bound."""
    diameter_dict = get_diameter_bins_dict(sensor_name)
    lower_bounds = [v[0] for v in diameter_dict["bounds"].values()]
    return lower_bounds


def get_diameter_bin_upper(sensor_name):
    """Get diameter bin upper bound."""
    diameter_dict = get_diameter_bins_dict(sensor_name)
    upper_bounds = [v[1] for v in diameter_dict["bounds"].values()]
    return upper_bounds


def get_diameter_bin_width(sensor_name):
    """Get diameter bin width."""
    diameter_dict = get_diameter_bins_dict(sensor_name)
    diameter_bin_width = list(diameter_dict["width"].values())
    return diameter_bin_width


def get_velocity_bin_center(sensor_name):
    """Get velocity bin center."""
    velocity_dict = get_velocity_bins_dict(sensor_name)
    velocity_bin_center = list(velocity_dict["center"].values())
    return velocity_bin_center


def get_velocity_bin_lower(sensor_name):
    """Get velocity bin lower bound."""
    velocity_dict = get_velocity_bins_dict(sensor_name)
    lower_bounds = [v[0] for v in velocity_dict["bounds"].values()]
    return lower_bounds


def get_velocity_bin_upper(sensor_name):
    """Get velocity bin upper bound."""
    velocity_dict = get_velocity_bins_dict(sensor_name)
    upper_bounds = [v[1] for v in velocity_dict["bounds"].values()]
    return upper_bounds


def get_velocity_bin_width(sensor_name):
    """Get velocity bin width."""
    velocity_dict = get_velocity_bins_dict(sensor_name)
    velocity_bin_width = list(velocity_dict["width"].values())
    return velocity_bin_width


def get_raw_field_nbins(sensor_name):
    diameter_dict = get_diameter_bins_dict(sensor_name)
    velocity_dict = get_velocity_bins_dict(sensor_name)
    n_d = len(diameter_dict["center"])
    n_v = len(velocity_dict["center"])
    nbins_dict = {
        "raw_drop_concentration": n_d,
        "raw_drop_average_velocity": n_v,
        "raw_drop_number": n_d * n_v,
    }
    return nbins_dict


# -----------------------------------------------------------------------------.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 24 20:30:51 2022

@author: ghiggi
"""
import os
import time
import glob 
import shutil
# Directory
from disdrodb.L0.io import (
    check_directories,
    get_campaign_name,
    create_directory_structure,
    get_L0A_dir, 
    get_L0A_fpath, 
    get_L0B_fpath,
    read_L0A_dataframe,
)
# Metadata 
from disdrodb.L0.metadata import read_metadata
# Standards 
from disdrodb.L0.check_standards import check_sensor_name, check_L0A_standards
# L0A_processing
from disdrodb.L0.L0A_processing import (
    get_file_list,
    read_L0A_raw_file_list,
    write_df_to_parquet,
)
# L0B_processing
from disdrodb.L0.L0B_processing import (
    create_L0B_from_L0A,
    write_L0B,
    create_summary_statistics,
)
# Logger 
from disdrodb.utils.logger import create_L0_logger, close_logger
from disdrodb.utils.logger import log_info, log_warning

#------------------------------------------------------------------.
# Consistency choice
# TODO: 
# - Add verbose and logs to disdrodb.io function !!!

#------------------------------------------------------------------.
def run_L0(
        # Arguments custom to each reader
        column_names,
        reader_kwargs,
        files_glob_pattern,
        df_sanitizer_fun, 
        raw_dir,
        processed_dir, 
        # Arguments designing the processing type
        l0a_processing=True,
        l0b_processing=True,
        keep_l0a=True,
        force=False,
        verbose=False,
        debugging_mode=False,
        lazy=True,
        single_netcdf=True,
        ):
    """Core function to process raw data files to L0A and L0B format. 
    
    Parameters
    ----------

    column_names : list 
        Column names to be assigned to the dataframe created after reading a raw file.
        This argument is allowed to contains column names not agreeing with the
        DISDRODB standard. However, it's important that non-standard columns are
        dropped during the application of the `df_sanitizer_fun` function.
    reader_kwargs : dict 
        Dictionary of arguments to be passed to `pd.read_csv` or `dd.read_csv` to 
        tailor the reading of the raw file. 
    files_glob_pattern: str
        It indicates the glob pattern to search for the raw data files within 
        the directory path raw_dir/data/<station_id>. 
    df_sanitizer_fun : function  
        It's a function detailing the custom preprocessing of each single raw data files 
        in order to meet the DISDRODB standards. 
        The function must have the following definition and return a dataframe: 
            def df_sanitizer_fun(df, lazy=False)
                # Import dask or pandas depending on lazy argument
                if lazy: 
                    import dask.dataframe as dd
                else: 
                    import pandas as dd
                # Custom processing 
                pass 
                # Return the dataframe agreeing with the DISDRODB standards
                return df 
    raw_dir : str
        Directory path of raw file for a specific campaign.
        The path should end with <campaign_name>.
        Example raw_dir: '<...>/disdrodb/data/raw/<campaign_name>'.
        The directory must have the following structure:
        - /data/<station_id>/<raw_files>
        - /metadata/<station_id>.json 
        Important points:
        - For each <station_id> there must be a corresponding JSON file in the metadata subfolder.
        - The <campaign_name> must semantically match between:
           - the raw_dir and processed_dir directory paths;
           - with the key 'campaign_name' within the metadata YAML files. 
        - The campaign_name are set to be UPPER CASE. 
    processed_dir : str
        Desired directory path for the processed L0A and L0B products. 
        The path should end with <campaign_name> and match the end of raw_dir.
        Example: '<...>/disdrodb/data/processed/<campaign_name>'.
    L0A_processing : bool
      Whether to launch processing to generate L0A Apache Parquet file(s) from raw data.
      The default is True.
    L0B_processing : bool
      Whether to launch processing to generate L0B netCDF4 file(s) from L0A data. 
      The default is True.
    keep_L0A : bool 
        Whether to keep the L0A files after having generated the L0B netCDF products.
        The default is False.
    force : bool
        If True, overwrite existing data into destination directories. 
        If False, raise an error if there are already data into destination directories. 
        The default is False.
    verbose : bool
        Whether to print detailed processing information into terminal. 
        The default is False.
    debugging_mode : bool
        If True, it reduces the amount of data to process.
        - For L0A processing, it processes just 3 raw data files.
        - For L0B processing, it takes a small subset of the L0A Apache Parquet dataframe.
        The default is False.
    lazy : bool
        Whether to perform processing lazily with dask. 
        If lazy=True, it employed dask.array and dask.dataframe.
        If lazy=False, it employed pandas.DataFrame and numpy.array.
        The default is True.
    single_netcdf : bool
        Whether to concatenate all raw files into a single L0B netCDF file.
        If single_netcdf=True, all raw files will be saved into a single L0B netCDF file.
        If single_netcdf=False, each raw file will be converted into the corresponding L0B netCDF file.
        The default is True.
    
    """
            
    t_i_script = time.time()
    #-------------------------------------------------------------------------.    
    # Initial directory checks 
    raw_dir, processed_dir = check_directories(raw_dir, processed_dir, force=force)
    
    # Retrieve campaign name 
    campaign_name = get_campaign_name(raw_dir)

    #-------------------------------------------------------------------------. 
    # Define logging settings
    logger = create_L0_logger(processed_dir, campaign_name)
            
    #-------------------------------------------------------------------------. 
    # Create directory structure 
    create_directory_structure(raw_dir, processed_dir)
                   
    #-------------------------------------------------------------------------. 
    #### Loop over station_id directory and process the files 
    list_stations_id = os.listdir(os.path.join(raw_dir, "data"))
    
    # station_id = list_stations_id[0]  
    for station_id in list_stations_id:
        #---------------------------------------------------------------------. 
        logger.info(f' - Processing of station_id {station_id} has started')
        #---------------------------------------------------------------------. 
        # Retrieve metadata 
        attrs = read_metadata(raw_dir=raw_dir,
                              station_id=station_id)
		# Retrieve sensor name
        sensor_name = attrs['sensor_name']
        check_sensor_name(sensor_name)
        
        #---------------------------------------------------------------------. 
        ######################## 
        #### L0A processing ####
        ######################## 
        if l0a_processing: 
            # Start L0 processing 
            t_i_station = time.time() 
            msg = " - L0A processing of station_id {} has started.".format(station_id)
            if verbose:
                print(msg)
            logger.info(msg)
            
            #-----------------------------------------------------------------.           
            #### - List files to process 
            glob_pattern = os.path.join("data", station_id, files_glob_pattern)
            file_list = get_file_list(raw_dir=raw_dir,
                                      glob_pattern=glob_pattern, 
                                      verbose=verbose, 
                                      debugging_mode=debugging_mode)
            
            #-----------------------------------------------------------------. 
            #### - If single_netcdf = True, ensure loop over all files only once   
            if single_netcdf: 
                 file_list = [file_list]
            
            #-----------------------------------------------------------------.
            #### - Loop over all files 
            # - It loops only once if single_netcdf=True
            for filepath in file_list:
                ##------------------------------------------------------.
                # Define file suffix
                if single_netcdf: 
                    file_suffix = ""
                else: 
                    # Get file name without file extensions 
                    t_i_file = time.time() 
                    file_suffix = os.path.basename(filepath).split(".")[0]  
                    logger.info(f"L0A processing of raw file {file_suffix} has started.")
                
                ##------------------------------------------------------.
                #### - Read all raw data files into a dataframe  
                df = read_L0A_raw_file_list(file_list=filepath,
                                           column_names=column_names,
                                           reader_kwargs=reader_kwargs,
                                           df_sanitizer_fun=df_sanitizer_fun,
                                           lazy=lazy,
                                           sensor_name=sensor_name,
                                           verbose=verbose)
            
                ##------------------------------------------------------.                                   
                #### - Write to Parquet                
                fpath = get_L0A_fpath(processed_dir, station_id, suffix=file_suffix)
                write_df_to_parquet(df=df,
                                    fpath=fpath,  
                                    force = force,
                                    verbose = verbose)
                
                ##------------------------------------------------------. 
                #### - Check L0 file respects the DISDRODB standards         
                check_L0A_standards(fpath=fpath, 
                                   sensor_name=sensor_name, 
                                   verbose=verbose)
                
                ##------------------------------------------------------.
                # Delete temp variables
                del df
                
                ##------------------------------------------------------.
                if not single_netcdf:
                    # End L0 processing for a single raw file
                    t_f = time.time() - t_i_file
                    msg = " - L0A processing of {} ended in {:.2f}s".format(file_suffix, t_f)
                    log_info(logger, msg, verbose)
                    
            ##------------------------------------------------------. 
            # End L0 processing for the station
            t_f = time.time() - t_i_station
            msg = " - L0A processing of station_id {} ended in {:.2f}s".format(station_id, t_f)
            log_info(logger, msg, verbose)
                 
        #------------------------------------------------------------------.
        ########################
        #### L0B processing ####
        ########################
        if l0b_processing: 
            # Start L1 processing 
            t_i = time.time() 
            msg = " - L0B processing of station_id {} has started.".format(station_id)
            if verbose:
                print(msg)
            logger.info(msg)
            ##----------------------------------------------------------------.
            # Get station L0A directory  
            L0A_dir_path = get_L0A_dir(processed_dir, station_id)
            file_list = glob.glob(os.path.join(L0A_dir_path,"*.parquet"))
            n_files = len(file_list)
            if n_files == 0: 
               msg = f"No L0A Apache Parquet file is available in {L0A_dir_path}. Run L0A processing first."
               logger.error(msg)
               raise ValueError(msg)
            
            ##----------------------------------------------------------------.
            # Checks for single_netcdf=True
            if single_netcdf :
                # Enclose into a list to loop over only once 
                file_list = [file_list] 
                if n_files != 1: 
                    msg = "If single_netcdf=True, DISDRODB would typically expect only a single L0A Apache Parquet file in {L0A_dir_path}."
                    log_warning(logger, msg, verbose)
            
            ##----------------------------------------------------------------.
            # Loop over all files 
            for filepath in file_list:
                ##------------------------------------------------------.
                # Define file suffix
                if single_netcdf: 
                    file_suffix = ""
                else: 
                    # Get file name without file extensions 
                    t_i_file = time.time() 
                    file_suffix = os.path.basename(filepath).split(".")[0]  
                    logger.info(f"L0A processing of raw file {file_suffix} has started.")
                ##------------------------------------------------------.    
                # Read L0A dataframes                
                df = read_L0A_dataframe(filepath, 
                                        lazy=lazy,
                                        verbose=verbose,
                                        debugging_mode=debugging_mode)
                  
                #-----------------------------------------------------------------.
                #### - Create xarray Dataset
                ds = create_L0B_from_L0A(df=df, attrs=attrs, lazy=lazy, verbose=verbose)
                    
                #-----------------------------------------------------------------.
                #### - Write L0B netCDF4 dataset
                fpath = get_L0B_fpath(processed_dir, station_id, suffix=file_suffix)
                write_L0B(ds, fpath=fpath, sensor_name=sensor_name)
                
                #-----------------------------------------------------------------.
                if not single_netcdf:
                    # End L0B processing for a single L0A file
                    t_f = time.time() - t_i_file
                    msg = " - L0B processing of {} ended in {:.2f}s".format(file_suffix, t_f)
                    log_info(logger, msg, verbose)
                    
            #-----------------------------------------------------------------.
            #### - Compute L0B summary statics (if single_netcdf=True)
            if single_netcdf:
                create_summary_statistics(ds, 
                                          processed_dir=processed_dir,
                                          station_id=station_id,
                                          sensor_name=sensor_name)
               
            #-----------------------------------------------------------------.
            # End L0B processing 
            t_f = time.time() - t_i
            msg = " - L0B processing of station_id {} ended in {:.2f}s".format(station_id, t_f)
            log_info(logger, msg, verbose)
 
            #-----------------------------------------------------------------.
        #---------------------------------------------------------------------.
    # Remove L0A directory if keep_L0A = False 
    if not keep_l0a: 
        shutil.rmtree(os.path.join(processed_dir, "L0A"))

    #-------------------------------------------------------------------------.
    # End of L0B processing for all stations
    t_f = time.time() - t_i_script
    msg = " - L0 processing of stations {} ended in {:.2f} minutes".format(list_stations_id, t_f/60)
    
    #-------------------------------------------------------------------------.
    # Final logs 
    logger.info('---')    
    msg = '### Script finish ###'
    log_info(logger, msg, verbose)
    close_logger(logger)
    


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = 'v0.0.4'
DESCRIPTION = "This package provides an API to query, filter and visualize disdrometer data."



# Setting up
setup(
    name="disdrodb",
    version=VERSION,
    author="Gionata Ghiggi",
    author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "click",
        "dask",
        "h5py",
        "netCDF4",
        "numpy",
        "pandas",
        "pyarrow",
        "PyYAML",
        "setuptools",
        "xarray"
        ],
    keywords=["python", "disdrometer"],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 18 12:20:46 2022

@author: ghiggi
"""

# TODO: adapt to new readers !!!!

def get_parser_cmd(
    parser_filepath,
    raw_dir,
    processed_dir,
    l0a_processing,
    l0b_processing,
    keep_l0a,
    force,
    verbose,
    debugging_mode,
    lazy,
    single_netcdf, 
):
    """Create command to launch parser processing from Terminal."""
    # parser_TICINO_2018.py [OPTIONS] <raw_dir> <processed_dir>
    cmd = "".join(
        [
            "python",
            " ",
            parser_filepath,
            " ",
            "--l0a_processing=",
            str(l0a_processing),
            " ",
            "--l0b_processing=",
            str(l0b_processing),
            " ",
            "--force=",
            str(force),
            " ",
            "--verbose=",
            str(verbose),
            " ",
            "--debugging_mode=",
            str(debugging_mode),
            " ",
            "--lazy=",
            str(lazy),
            " ",
            "--keep_l0a=",
            str(keep_l0a),
            " ",
            "--single_netcdf=",
            str(single_netcdf),
            " ",
            raw_dir,
            " ",
            processed_dir,
        ]
    )
    return cmd



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 23 18:44:52 2022

@author: ghiggi
"""
import os
import yaml 
import logging
logger = logging.getLogger(__name__)


def _write_issue_timestamps_docs(f): 
    """Provide template for timestamps.yml """
    # TODO: Kim adapt to desired format 
    f.write("# This file is used to store dates to drop by the parser, the time format used is the isoformat (YYYY-mm-dd HH:MM:SS). \n")
    f.write("# timestamp: list of timestamps \n")
    f.write("# time_period: list of list ranges of dates \n")
    f.write("# Example: \n")
    f.write("# timestamp: ['2018-12-07 14:15','2018-12-07 14:17','2018-12-07 14:19', '2018-12-07 14:25'] \n")
    f.write("# time_period: [['2018-08-01 12:00:00', '2018-08-01 14:00:00'], \n")
    f.write("#               ['2018-08-01 15:44:30', '2018-08-01 15:59:31'], \n")
    f.write("#               ['2018-08-02 12:44:30', '2018-08-02 12:59:31']] \n")
    
def create_issue_yml(fpath, timestamp=None, time_period=None):
    """Write timestamps.yml """
    logger.info(f'Creating issue YAML file at {fpath}')
    with open(fpath, "w") as f:
        # Write template for timestamps.yml
        _write_issue_timestamps_docs(f)
        
        # Write timestamp if provided 
        # TODO: Kim adapt to desired format
        if timestamp is not None: 
            timestamp_dict = {"timestamp": timestamp}
            yaml.dump(timestamp_dict, f,
                      default_flow_style=None)
            
        # Write timestamp if provided 
        # TODO: Kim adapt to desired format
        if time_period is not None:
            time_period_dict = {"time_period": time_period}
            yaml.dump(time_period_dict, f, 
                      default_flow_style=None)

def read_issue(raw_dir, station_id):
    """Read YAML issue file."""
    issue_fpath = os.path.join(raw_dir, "issue", station_id + ".yml")
    with open(issue_fpath, "r") as f:
        issue_dict = yaml.safe_load(f)
    # issue_dict = check_issue_compliance(issue_dict)
    return issue_dict

def check_issue_compliance(fpath):
    pass
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------.
# Copyright (c) 2021-2022 DISDRODB developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------.

import os
import yaml


def get_attrs_standards():
    # TO REMOVE
    # - station_number
    # - disdrodb_id 
    # - temporal_resolution
    
    # TO ADD
    # - comments
    # - acknoledgements
    # - license   
    
    # sensor_wavelegth --> sensor_wavelength
    
    list_attrs = [  # Description
        "title",
        "description",
        "source",
        "history",
        "conventions",
        "campaign_name",
        "project_name",
        # Location
        "station_id",
        "station_name",
        # "station_number",TODO: REMOVE 
        "location",
        "country",
        "continent",
        "latitude",
        "longitude",
        "altitude",
        "crs",
        "proj4_string",
        "EPSG",
        "latitude_unit",
        "longitude_unit",
        "altitude_unit",
        # Sensor info
        "sensor_name",
        "sensor_long_name",
        "sensor_wavelength",
        "sensor_serial_number",
        "firmware_iop",
        "firmware_dsp",
        "firmware_version",
        "sensor_beam_width",
        "sensor_nominal_width",
        # "temporal_resolution",# TODO REMOVE
        "measurement_interval",
        # Attribution
        "contributors",
        "authors",
        "institution",
        "references",
        "documentation",
        "website",
        "source_repository",
        "doi",
        "contact",
        "contact_information",
        # Source datatype
        "source_data_format",
        # DISDRO DB attrs
        "obs_type",
    ]
    attrs = {key: "" for key in list_attrs}
    # TODO: temporary attributes for EPFL development
    # attrs["sensor_name"] = "OTT_Parsivel"
    attrs["latitude"] = -9999
    attrs["longitude"] = -9999
    attrs["altitude"] = -9999
    # attrs["institution"] = "Laboratoire de Teledetection Environnementale -  Ecole Polytechnique Federale de Lausanne"
    # attrs["sensor_long_name"] = "OTT Hydromet Parsivel"
    # attrs["contact_information"] = "http://lte.epfl.ch"

    # Defaults attributes
    attrs["latitude_unit"] = "DegreesNorth"
    attrs["longitude_unit"] = "DegreesEast"
    attrs["altitude_unit"] = "MetersAboveSeaLevel"
    attrs["crs"] = "WGS84"
    attrs["EPSG"] = 4326
    attrs["proj4_string"] = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"

    return attrs

# TODO: create_metadata_yml ? Decide similar pattern to create_issue<_yml>
def create_metadata(fpath):
    """Create default YAML metadata file."""
    attrs = get_attrs_standards()
    with open(fpath, "w") as f:
        yaml.dump(attrs, f, sort_keys=False)


def read_metadata(raw_dir, station_id):
    """Read YAML metadata file."""
    metadata_fpath = os.path.join(raw_dir, "metadata", station_id + ".yml")
    with open(metadata_fpath, "r") as f:
        attrs = yaml.safe_load(f)
    return attrs


def check_metadata_compliance(raw_dir):
    """Check YAML metadata files compliance."""
    # TODO: MISSING CHECKS
    # - CHECK NO MISSING IMPORTANT METADATA
    # - CHECK NO ADDITIONAL METADATA OTHER THAN STANDARDS
    # - CHECK VALUE VALIDITY OF KEYS REQUIRED FOR PROCESSING
    # check_sensor_name(sensor_name=sensor_name)
    pass
    return


# def write_metadata(attrs, fpath):
#     """Write dictionary to YAML file."""
#     with open(fpath, "w") as f:
#         yaml.dump(attrs, f, sort_keys=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan  2 14:56:38 2022

@author: ghiggi
"""
import numpy as np
import pandas as pd
from disdrodb.L0.standards import get_L0A_dtype
from disdrodb.L0.check_standards import (
    get_field_nchar_dict,
    get_field_ndigits_dict,
    get_field_ndigits_decimals_dict,
    get_field_ndigits_natural_dict,
)

def check_column_names(column_names, sensor_name): 
    "Checks that the columnn names respects DISDRODB standards."
    if isinstance(column_names, list):
        raise TypeError("'column_names' must be a list of strings.")
    # Get valid columns 
    dtype_dict = get_L0_dtype(sensor_name)
    valid_columns = list(dtype_dict)
    valid_columns = valid_columns + ['time']
    # --------------------------------------------
    # Create name sets
    column_names = set(column_names)
    valid_columns = set(valid_columns)  
    # --------------------------------------------
    # Raise warning if there are columns not respecting DISDRODB standards 
    unvalid_columns = list(column_names.difference(valid_columns))
    if len(unvalid_columns) > 0:
        print("The following columns do no met the DISDRODB standards: {unvalid_columns}.")
        print("Please remove such columns within the df_sanitizer_fun")
    # --------------------------------------------
    # Check time column is present 
    if 'time' not in column_names:
        print("Please be sure to create the 'time' column within the df_sanitizer_fun.")
        print("The 'time' column must be datetime with resolution in seconds (dtype='M8[s]').")
    # --------------------------------------------
    return None

    def check_L0_column_names(x):
        # TODO: 
        # check_L0_column_names(column_names, sensor_name) 
        # --> Move in for loop 
        # --> Print message with columns to be drop in df_sanitizer 
        # --> Print message of columns to be derived in df_sanitizer (i.e. time)
        pass
    
    
def print_df_first_n_rows(df, n=5, column_names=True):
    columns = list(df.columns)
    for i in range(len(df.columns)):
        if column_names:
            print(" - Column", i, "(", columns[i], "):")
        else:
            print(" - Column", i, ":")
        print("     ", df.iloc[0 : (n + 1), i].values)
    return None


def print_df_random_n_rows(df, n=5, column_names=True):
    columns = list(df.columns)
    df = df.copy()
    df = df.sample(n=n)
    for i in range(len(df.columns)):
        if column_names:
            print(" - Column", i, "(", columns[i], "):")
        else:
            print(" - Column", i, ":")
        print("     ", df.iloc[0 : (n + 1), i].values)
    return None


def print_df_column_names(df):
    for i, column in enumerate(df.columns):
        print(" - Column", i, ":", column)
    return None


def print_valid_L0_column_names(sensor_name):
    print(list(get_L0_dtype(sensor_name)))
    return None


def _check_valid_column_index(column_idx, n_columns):
    if column_idx > (n_columns - 1):
        raise ValueError("'column_idx' must be between 0 and {}".format(n_columns - 1))
    if column_idx < 0:
        raise ValueError("'column_idx' must be between 0 and {}".format(n_columns - 1))


def _check_columns_indices(column_indices, n_columns):
    if not isinstance(column_indices, (int, list, type(None), slice)):
        raise TypeError(
            "'column_indices' must be an integer, a list of integers, or None."
        )
    if column_indices is None:
        column_indices = list(range(0, n_columns))
    if isinstance(column_indices, slice):
        start = column_indices.start
        stop = column_indices.stop
        step = column_indices.step
        step = 1 if step is None else step
        column_indices = list(range(start, stop, step))
    if isinstance(column_indices, list):
        [_check_valid_column_index(idx, n_columns) for idx in column_indices]
    if isinstance(column_indices, int):
        _check_valid_column_index(column_indices, n_columns)
        column_indices = [column_indices]
    return column_indices


def print_df_columns_unique_values(df, column_indices=None, column_names=True):
    # Retrieve column names
    columns = list(df.columns)
    n_columns = len(columns)
    # Checks for printing specific columns only
    column_indices = _check_columns_indices(column_indices, n_columns)
    columns = [columns[idx] for idx in column_indices]
    # Printing
    for i, column in zip(column_indices, columns):
        if column_names:
            print(" - Column", i, "(", column, "):")
        else:
            print(" - Column", i, ":")
        print("     ", sorted(df[column].unique().tolist()))
    return None


def get_df_columns_unique_values_dict(df, column_indices=None, column_names=True):
    """Create a dictionary {column: unique values}"""
    # Retrieve column names
    columns = list(df.columns)
    n_columns = len(columns)
    # Checks for printing specific columns only
    column_indices = _check_columns_indices(column_indices, n_columns)
    columns = [columns[idx] for idx in column_indices]
    # Create dictionary
    d = {}
    for i, column in zip(column_indices, columns):
        if column_names:
            key = column
        else:
            key = "Column " + str(i)
        d[key] = sorted(df[column].unique().tolist())
    # Return
    return d


def print_df_summary_stats(df, column_indices=None, column_names=True):
    # Define columns of interest
    columns = df.columns
    n_columns = len(columns)
    column_indices = _check_columns_indices(column_indices, n_columns)
    columns_of_interest = [columns[idx] for idx in column_indices]
    # Remove columns of dtype object or string
    indices_to_remove = np.where((df.dtypes == type(object)) | (df.dtypes == str))
    indices = np.arange(0, len(df.columns))
    indices = indices[np.isin(indices, indices_to_remove, invert=True)]
    columns = df.columns[indices]
    if len(columns) == 0:
        raise ValueError("No numeric columns in the dataframe.")
    # Select only columns of interest
    idx_of_interest = np.where(np.isin(columns, columns_of_interest))
    if len(idx_of_interest) == 0:
        raise ValueError("No numeric columns at the specified column_indices.")
    columns = columns[idx_of_interest]
    indices = indices[idx_of_interest]
    # Compute summary stats
    summary_stats = ["mean", "min", "25%", "50%", "75%", "max"]
    df_summary = df.describe()
    df_summary = df_summary.loc[summary_stats]
    # Print summary stats
    for i, column in zip(indices, columns):
        # Printing

        if column_names:
            print(" - Column", i, "(", column, "):")
        else:
            print(" - Column", i, ":")
        tmp_df = df_summary[[column]]
        tmp_df.columns = [""]
        print("     ", tmp_df)
    return None


def print_df_with_any_nan_rows(df):
    df_bool_is_nan = df.isnull()
    idx_nan_rows = df_bool_is_nan.any(axis=1)
    df_nan_rows = df.loc[idx_nan_rows]
    print_df_first_n_rows(df_nan_rows, n=len(df_nan_rows))


####--------------------------------------------------------------------------.
#### Character checks
def arr_has_constant_nchar(arr):
    arr = np.asarray(arr)
    if arr.dtype.char not in ["O", "U"]:
        raise TypeError("Expecting object (O) or string (U) dtype.")
    if arr.dtype.char == "O":
        arr = arr.astype(str)
    if arr.dtype.char != "U":
        raise TypeError("Expecting string (U) dtype.")
    # Get number of characters (include .)
    str_nchars = np.char.str_len(arr)
    str_nchars_unique = np.unique(str_nchars)
    if len(str_nchars_unique) != 1:
        return False  # raise ValueError("Non-unique string length !")
    else:
        return True


def str_is_number(string):
    try:
        float(string)
        return True
    except ValueError:
        return False


def str_is_not_number(string):
    return not str_is_number(string)


def str_is_integer(string):
    try:
        int(string)
        return True
    except ValueError:
        return False


def str_has_decimal_digits(string):
    if len(string.split(".")) == 2:
        return True
    else:
        return False


def get_decimal_ndigits(string):
    if str_has_decimal_digits(string):
        return len(string.split(".")[1])
    else:
        return 0


def get_natural_ndigits(string):
    if str_is_integer(string):
        return len(string)
    if str_has_decimal_digits(string):
        return len(string.split(".")[0])
    else:
        return 0


def get_ndigits(string):
    if str_is_not_number(string):
        return 0
    if str_has_decimal_digits(string):
        return len(string) - 1  # remove .
    else:
        return len(string)


def get_nchar(string):
    return len(string)


def get_possible_keys(dict_options, desired_value):
    list_key_match = []
    for k, v in dict_options.items():
        if v == desired_value:
            list_key_match.append(k)
    set_key_match = set(list_key_match)
    return set_key_match


def search_possible_columns(string, sensor_name):
    dict_digits = get_field_ndigits_dict(sensor_name)
    dict_nchar_digits = get_field_nchar_dict(sensor_name)
    dict_decimal_digits = get_field_ndigits_decimals_dict(sensor_name)
    dict_natural_digits = get_field_ndigits_natural_dict(sensor_name)
    set_digits = get_possible_keys(dict_digits, get_ndigits(string))
    set_nchar = get_possible_keys(dict_nchar_digits, get_nchar(string))
    set_decimals = get_possible_keys(dict_decimal_digits, get_decimal_ndigits(string))
    set_natural = get_possible_keys(dict_natural_digits, get_natural_ndigits(string))
    possible_keys = set_digits.intersection(set_nchar, set_decimals, set_natural)
    possible_keys = list(possible_keys)
    return possible_keys


def infer_df_str_column_names(df, sensor_name, row_idx=1):
    dict_possible_columns = {}
    for i, column in enumerate(df.columns):
        print(i)
        # Get string array
        arr = df.iloc[:, i]
        arr = np.asarray(arr).astype(str)
        if not arr_has_constant_nchar(arr):
            print("Column", i, "has non-unique number of characters")
            continue
        # Subset a single string
        string = arr[row_idx]
        possible_columns = search_possible_columns(string, sensor_name=sensor_name)
        dict_possible_columns[i] = possible_columns
    return dict_possible_columns
